######################################################
class input_profiles_adaptor(OMFITinterfaceAdaptor):
    type = 'input_profiles'

    @staticmethod
    def identify(data):
        import numpy as np
        import numpy
        printd('Attempting to identify {:} type...'.format('input_profiles'))
        if isinstance(data, dict) and np.all([k in data for k in ['Ti_1', 'ni_1']]):
            printd(' > Looks like it is {:} type!'.format('input_profiles'))
            return True
        else:
            printd(' > It does not seem to be {:} type.'.format('input_profiles'))

    def definitions(self, key, data):
        if self.register(key, 'time', 'Time', 's', ''):
            return 0

        # -------------------------
        # Geometry
        # -------------------------
        if self.register(key, 'rho', 'Normalized minor radius', '1', 'r'):
            return data['rho']

        if self.register(key, 'psi', 'Poloidal flux', 'V*s/rad', 'r'):
            return data['polflux']

        if self.register(key, ['psin', 'psi_n'], 'Normalized poloidal flux', '1', 'r'):
            return (self['psi'] - min(self['psi'])) / (max(self['psi']) - min(self['psi']))

        # -------------------------
        # Energy, pressure and power
        # -------------------------
        if self.register(key, ['ptot', 'pressure'], 'Total pressure', 'N/m^2', 'r'):
            return data['ptot']

        # -------------------------
        # Temperature and densities
        # -------------------------
        if self.register(key, ['ne', 'electron density'], 'Electron density', '1/m^3', 'r'):
            return data['ne'] * 1E19

        # -------------------------
        # Other plasma parameters
        # -------------------------
        if self.register(key, 'Zeff', 'Effective impurity charge', 'amu', 'r'):
            return data['z_eff']


######################################################
class ONETWO_statefile_adaptor(OMFITinterfaceAdaptor):
    type = 'ONETWO_statefile'

    @staticmethod
    def identify(data):
        from omfit_classes.omfit_base import OMFITcollection
        printd('Attempting to identify {:} type...'.format('ONETWO_statefile'))
        if isinstance(data, dict) and not isinstance(data, OMFITcollection) and np.all([k in data for k in ['press', 'rho_grid']]):
            printd(' > Looks like it is {:} type!'.format('ONETWO_statefile'))
            return True
        else:
            printd(' > It does not seem to be {:} type.'.format('ONETWO_statefile'))

    def definitions(self, key, data):
        import numpy as np
        import numpy
        printd('ONETWO_statefile definitions...')

        def interp_npsi_vars_to_rho(val):
            rho = data['rho_grid']['data']
            rho_mhd = np.flipud(data['rho_mhd_gridnpsi']['data'])
            val = np.flipud(val)
            return interp1e(rho_mhd, val, kind='linear')(rho)

        if self.register(key, 'time', 'Time', 's', ''):
            return data['time']['data']

        # -------------------------
        # Geometry
        # -------------------------
        if self.register(key, 'rho', 'Normalized minor radius', '1', 'r'):
            return np.linspace(0, 1, len(data['rho_grid']['data']))

        if self.register(key, 'rho_m', 'Minor radius', 'm', 'r'):
            return data['rho_grid']['data']

        if self.register(key, 'a', 'Plasma midplane minor radius', 'm', ''):
            return interp_npsi_vars_to_rho(data['rminavnpsi']['data'])[-1]

        if self.register(key, 'psi', 'Poloidal flux', 'V*s/rad', 'r'):
            return data['psir_grid']['data']

        if self.register(key, ['psin', 'psi_n'], 'Normalized poloidal flux', '1', 'r'):
            return (self['psi'] - min(self['psi'])) / (max(self['psi']) - min(self['psi']))

        if self.register(key, 'r_min', 'Geometric midplane minor radius', 'm', 'r'):
            return interp_npsi_vars_to_rho(data['rminavnpsi']['data'])

        if self.register(key, ['r_geo', 'r_maj'], 'Geometric midplane major radius', 'm', 'r'):
            return interp_npsi_vars_to_rho(data['rmajavnpsi']['data'])

        if self.register(key, ['rout', 'r_out'], 'Geometric midplane outer major radius', 'm', 'r'):
            return interp_npsi_vars_to_rho(data['rmajavnpsi']['data'] + data['rminavnpsi']['data'])

        if self.register(key, ['R0', 'vacuum major radius'], 'Major radius of vacuum BT0 reference == R0', 'm', ''):
            return data['rmajor']['data']

        if self.register(key, '1/R', '<1/R>', '1/m', 'r'):
            return interp_npsi_vars_to_rho(data['ravginpsi']['data'])

        if self.register(key, 'kappa', 'Elongation', '1', 'r'):
            return interp_npsi_vars_to_rho(data['elongxnpsi']['data'])

        if self.register(key, 'delta', 'Triangularity', '1', 'r'):
            return interp_npsi_vars_to_rho(0.5 * data['triangnpsi_l']['data'] + 0.5 * data['triangnpsi_u']['data'])

        if self.register(key, ['Acx', 'plasma cross section'], 'Total plasma cross section', 'm^2', ''):
            return data['areao']['data']

        if self.register(key, 'area', 'Cross sectional area enclosed by each flux surface', 'm^2', 'r'):
            return data['sfareanrho']['data']

        if self.register(key, 'darea', 'Gradient of cross sectional area enclosed by each flux surface', 'm^2', 'rt'):
            return gradient(self['area'])

        if self.register(key, ['vol', 'V', 'volume'], 'Volume', 'm^3', 'r'):
            integrand = 4 * np.pi ** 2 * self['R0'] * self['rho_m'] * data['hcap']['data']
            return scipy.integrate.cumtrapz(integrand, self['rho_m'], initial=0)

        if self.register(key, ['dvol', 'dV/drho', 'differential volume'], 'Differential volume', 'm^3', 'r'):
            return deriv(self['rho'], self['volume'])

        if self.register(key, 'pprime', 'dp/dpsi', 'N/m^2', 'r'):
            return deriv(self['psi'], self['pressure'])

        # -------------------------
        # Fields
        # -------------------------
        if self.register(key, ['BT0', 'vacuum toroidal field'], 'Vacuum BT0 at R0', 'T', ''):
            return data['btor']['data']

        if self.register(key, ['Btot_out', 'total B field out', 'B magnitude out', 'btot_out'], 'B field magnitude at outboard midplane', 'T', 'r'):
            return data['btotrmaj']['data']

        if self.register(key, ['f', 'fpol', 'FPOL', 'I_psi'], 'Bz * R (including dia/paramagnetism)', 'm*T', 'r'):  # f of psi = R * Bt
            return data['fpsinrho']['data']

        if self.register(key, 'ffprime', 'f*df/dpsi', '(m*T)^2/(V*s/rad)', 'r'):
            return data['ffprim']['data']

        if self.register(key, ['Bp_avg', 'Bp_fsa', 'BPOL_AVG', 'Bpol_fsa'], 'Flux surface average of poloidal B field', 'T', 'r'):
            return data['bp']['data']

        ##        if self.register(key,['Btot_avg','total B field avg','B magnitude avg','btot_avg','btot_fsa'],'B field magnitude, flux surface average, tesla','T','r'):
        ##            bp_fsa=self['Bp_avg']
        ##            bt_fsa=
        ##            btotavg=np.sqrt(bp_fsa**2+bt_fsa**2)
        ##            return btotavg

        # -------------------------
        # Current
        # -------------------------
        if self.register(key, ['Ip', 'total current'], 'Total plasma current ', 'A', ''):
            return data['tot_cur']['data']

        if self.register(key, ['jtot', 'current'], 'Current density <Jtor R0/R>', 'A/m^2', 'r'):
            return data['curden']['data']

        if self.register(key, ['jpar', 'j//', 'parallel current'], 'Parallel current density <J.B/Bt0>', 'A/m^2', 'r'):
            return data['curpar']['data']

        if self.register(key, ['joh', 'johm', 'ohmic current'], 'Ohmic current density <Jtor R0/R>', 'A/m^2', 'r'):
            return data['curohm']['data']

        if self.register(key, ['jboot', 'bootstrap current'], 'Bootstrap current density <Jtor R0/R>', 'A/m^2', 'r'):
            return data['curboot']['data']

        if self.register(key, ['jb', 'beam current'], 'Beam current density <Jtor R0/R>', 'A/m^2', 'r'):
            return data['curbeam']['data']

        if self.register(key, ['jrf', 'rf current'], 'RF current density <Jtor R0/R>', 'A/m^2', 'r'):
            return data['currf']['data']

        if self.register(key, ['q_abs', 'safety factor magnitude'], 'Safety factor magnitude', '', 'r'):
            return data['q_value']['data']

        # -------------------------
        # Energy, pressure and power
        # -------------------------
        if self.register(key, ['betat'], 'Toroidal beta', '', ''):
            return data['beta']['data']

        if self.register(key, ['betap'], 'Poloidal beta', '', ''):
            return data['betap']['data']

        if self.register(key, ['betan'], 'Normalized beta', '', ''):
            bt = self['BT0'] * self['R0'] / self['r_geo'][-1]
            return abs(self['betat'] / (self['Ip'] / 1E6 / self['a'] / bt))

        if self.register(key, ['Wtot', 'energy'], 'Total energy', 'J/m^3', 'r'):
            return self['pressure'] * 3 / 2.

        if self.register(key, ['ptot', 'pressure'], 'Total pressure', 'N/m^2', 'r'):
            return data['press']['data']

        if self.register(key, ['Wbeam', 'beam energy'], 'Beam energy density', 'J/m^3', 'r'):
            return data['wbeam']['data'] * 1.60218e-16  # Conversion from keV/m^3 to J/m^3=N/m^2

        if self.register(key, ['pbeam', 'beam pressure'], 'Beam pressure', 'N/m^2', 'r'):
            return self['beam energy'] * 2 / 3.

        # -------------------------
        # Temperature and densities
        # -------------------------
        if self.register(key, ['Te', 'electron temperature'], 'Electron temperature', 'keV', 'r'):
            return data['Te']['data']

        if self.register(key, ['ne', 'electron density'], 'Electron density', '1/m^3', 'r'):
            return data['ene']['data']

        if self.register(key, ['Ti', 'ion temperature'], 'Ion temperature', 'keV', 'ir'):
            return np.array([data['Ti']['data']] * len(self['ion species']))

        if self.register(key, 'ion species', 'Ion species', '1', 'i'):
            return data['enion']['long_name'].split(':')[1].strip().split()  # both primary and impurity ions

        if self.register(key, ['ni', 'ion densities'], 'Ion densities', '1/m^3', 'ir'):
            return data['enion']['data']  # includes both primary and impurity ions

        if self.register(key, 'beam species', 'Beam ion species', '1', 'b'):
            return data['enbeam']['long_name'].split(':')[1].strip().split()

        if self.register(key, ['nb', 'beam densities'], 'Beam ion densities', '1/m^3', 'br'):
            return data['enbeam']['data']  # includes all beam species

        # -------------------------
        # Other plasma parameters
        # -------------------------
        if self.register(key, 'Zeff', 'Effective impurity charge', 'amu', 'r'):
            return data['zeff']['data']

        if self.register(key, ['Z', 'Zi', 'charge state'], 'Ion charge state', 'amu', 'i'):
            result = data['z']['data'][:, 0]
            printd('  >> zi: np.shape(result) = ', np.shape(result))
            return result

        if self.register(key, ['ftrap', 'trapfrac', 'trapped fraction', 'ncft'], 'Trapped particle fraction', '', 'r'):
            result = data['ftrap']['data']
            printd('  >> ftrap: np.shape(result) = ', np.shape(result))
            return result


######################################################
class ONETWO_MDS_adaptor(OMFITinterfaceAdaptor):
    type = 'ONETWO_MDS'

    @staticmethod
    def identify(data):
        printd('Attempting to identify {:} type...'.format('ONETWO_MDS'))
        if 'EQU' in data and data['EQU']['ANGROT'].data() is not None:
            printd(' > Looks like it is {:} type!'.format('ONETWO_MDS'))
            return True
        else:
            printd(' > It does not seem to be {:} type.'.format('ONETWO_MDS'))

    def definitions(self, key, data):
        import numpy as np
        import numpy
        printd('ONETWO_MDS definitions...')
        if self.register(key, 'device', 'Device', '1', ''):
            return 'D3D'

        if self.register(key, 'shot', 'Shot', '1', ''):
            return int(data['SHOT'].data())

        if self.register(key, 'time', 'Time', 's', 't'):
            return data['GOOD_TIMES'].data()

        # -------------------------
        # Geometry
        # -------------------------
        if self.register(key, 'a', 'Plasma midplane minor radius', 'm', 't'):
            if 'RMIN_MILLER' in data['EQU']:
                return data['EQU']['RMIN_MILLER'].data()[-1, :]
            else:
                return data['EQU']['RMINOR'].data()[-1, :]

        if self.register(key, 'psi', 'Poloidal flux', 'V*s/rad', 'rt'):
            return data['EQU']['POL_FLUX'].data()

        if self.register(key, ['psin', 'psi_n'], 'Normalized poloidal flux', '1', 'rt'):
            return (self['psi'] - np.min(self['psi'], axis=0)[np.newaxis, :]) / (np.max(self['psi'], axis=0)[np.newaxis, :] - np.min(self['psi'], axis=0)[np.newaxis, :])

        if self.register(key, 'r_min', 'Geometric midplane minor radius', 'm', 'rt'):
            return self['a']

        if self.register(key, ['r_geo', 'r_maj'], 'Geometric midplane major radius', 'm', 'rt'):
            return data['EQU']['RMAJOR'].data()

        if self.register(key, ['rout', 'r_out'], 'Geometric midplane outer major radius', 'm', 'rt'):
            return data['EQU']['RMAJOR'].data() + self['a']

        if self.register(key, ['R0', 'vacuum major radius'], 'Major radius of vacuum BT0 reference == R0', 'm', ''):
            specs = omfit_classes.utils_fusion.device_specs(tolist(self['device'])[0])
            if 'R0' in list(specs.keys()) and specs['R0'] is not None:
                return specs['R0']
            else:
                raise Exception('Plasma vacuum major radius is not defined for ' + str(self['device']))

        if self.register(key, ['BT0', 'vacuum toroidal field'], 'Vacuum BT0 at R0', 'T', 't'):
            return data['EQU']['BTOR'].data()

        if self.register(key, 'rho_m', 'Minor radius', 'm', 'rt'):
            return data['EQU']['RHOM'].data()  # same as data['EQU']['RMINOR'].data()

        if self.register(key, 'rho', 'Normalized minor radius', '1', 'r'):
            return data['EQU']['RHO'].data()

        # if self.register(key,'1/R','1/R','1/m','rt'):
        #    return ???

        if self.register(key, 'kappa', 'Elongation', '1', 'rt'):
            return data['EQU']['KAPPAR'].data()

        if self.register(key, 'delta', 'Triangularity', '1', 'rt'):
            if 'DELTAR_L' in data['EQU'] and 'DELTAR_U' in data['EQU']:
                return 0.5 * data['EQU']['DELTAR_L'].data() + 0.5 * data['EQU']['DELTAR_U'].data()
            else:
                return data['EQU']['DELTAR'].data()

        if self.register(key, ['Acx', 'plasma cross section'], 'Total plasma cross section', 'm^2', 't'):
            return np.squeeze(data['EQU']['SURF_AREA'].data()[-1, :])

        if self.register(key, ['vol', 'V', 'volume'], 'Volume', 'm^3', 'rt'):
            return data['EQU']['VOL'].data()

        if self.register(key, ['dvol', 'dV/drho', 'differential volume'], 'Differential volume', 'm^3', 'rt'):
            return gradient(self['volume'])[0] / gradient(self.get('rho', 'rt'))[:, np.newaxis]

        if self.register(key, 'pprime', 'dp/dpsi', 'N/m^2', 'rt'):
            return gradient(self['pressure'])[0] / gradient(self['psi'])[0]

        # if self.register(key,'ffprime','f*df/dpsi','(m*T)^2/(V*s/rad)','r'):
        #     return ???

        # -------------------------
        # Current
        # -------------------------
        if self.register(key, ['Ip', 'total current'], 'Total plasma current ', 'A', 't'):
            return data['EQU']['IP'].data()

        if self.register(key, ['q_abs', 'safety factor magnitude'], 'Safety factor magnitude', '', 'rt'):
            return abs(data['EQU']['Q'].data())

        if self.register(key, ['jtot', 'current'], 'Current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return data['EQU']['J_EQUIL'].data()

        # if self.register(key,['jpar','j//','parallel current'],'Parallel current density <J.B/Bt0>','A/m^2','r'):
        #    return ???

        if self.register(key, ['joh', 'johm', 'ohmic current'], 'Ohmic current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return data['EQU']['J_OHM'].data()

        if self.register(key, ['jboot', 'bootstrap current'], 'Bootstrap current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return data['HCD']['J_NI'].data() - data['HCD']['J_CD'].data()

        if self.register(key, ['jb', 'beam current'], 'Beam current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return data['NBI']['J_NBCD'].data()

        if self.register(key, ['jrf', 'rf current'], 'RF current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return data['HCD']['J_CD'].data() - data['NBI']['J_NBCD'].data()

        # -------------------------
        # Energy, pressure and power
        # -------------------------
        if self.register(key, ['betat'], 'Toroidal beta', '', 't'):
            return data['EQU']['BETA'].data()

        if self.register(key, ['betap'], 'Poloidal beta', '', 't'):
            return data['EQU']['BETAP'].data()

        if self.register(key, ['betan'], 'Normalized beta', '', 't'):
            bt = self['BT0'] * self['R0'] / self['r_geo'][-1, :]
            return abs(self['betat'] / (self['Ip'] / 1E6 / self['a'] / bt))

        if self.register(key, ['Wtot', 'energy'], 'Total energy', 'J/m^3', 'rt'):
            return data['EQU']['PRESS_KIN'].data() * 3 / 2.

        if self.register(key, ['ptot', 'pressure'], 'Total pressure', 'N/m^2', 'rt'):
            return data['EQU']['PRESS_KIN'].data()

        if self.register(key, ['Wbeam', 'beam energy'], 'Beam energy density', 'J/m^3', 'rt'):
            return data['NBI']['WBEAM'].data()  # Already in J/m^3=N/m^2

        if self.register(key, ['pbeam', 'beam pressure'], 'Beam pressure', 'N/m^2', 'rt'):
            return data['EQU']['PRESS_BEAM'].data()

        # -------------------------
        # Temperature and densities
        # -------------------------
        if self.register(key, ['Te', 'electron temperature'], 'Electron temperature', 'keV', 'rt'):
            return data['EQU']['T_E'].data()

        if self.register(key, ['ne', 'electron density'], 'Electron density', '1/m^3', 'rt'):
            return data['EQU']['DENS_E'].data()

        if self.register(key, 'ion species', 'Ion species', '1', 'i'):
            return data['INPUTS']['NAMELIS1']['NAMEP'].data()[0][0].strip("'").split(',') + data['INPUTS']['NAMELIS1']['NAMEI'].data()[0][0].strip("'").split(',')

        if self.register(key, ['Ti', 'ion temperature'], 'Ion temperature', 'keV', 'irt'):
            return np.array([data['EQU']['T_I'].data()] * len(self['ion species']))

        if self.register(key, ['ni', 'ion densities'], 'Ion densities', '1/m^3', 'irt'):
            nion = []
            nion.append(data['EQU']['DENS_I'].data())
            nion.append(data['EQU']['DENS_IMP'].data())
            return np.array(nion)

        if self.register(key, 'beam species', 'Beam ion species', '1', 'b'):
            return data['INPUTS']['NAMELIS2']['NAMEB'].data()[0][0].strip("'").split(',')

        if self.register(key, ['nb', 'beam densities'], 'Beam ion densities', '1/m^3', 'brt'):
            return np.array([data['NBI']['ENBEAM'].data()] * len(self['beam species']))

        if self.register(key, ['Tb', 'beam temperature'], 'Beam ion temperature', 'keV', 'brt'):
            return np.array([data['NBI']['TBEAM'].data()] * len(self['beam species']))  # includes all beam species

        # -------------------------
        # Other plasma parameters
        # -------------------------
        if self.register(key, 'Zeff', 'Effective impurity charge', 'amu', 'rt'):
            return data['EQU']['Z_EFF'].data()

        # -------------------------
        # Diffusivities and fluxes
        # -------------------------
        if self.register(key, ['chie', 'electron heat diffisivity'], 'Electron heat diffisivity', '1/m^2/s', 'rt'):
            return data['DIFF']['CHI_E'].data()

        if self.register(key, ['chii', 'ion heat diffisivity'], 'Ion heat diffisivity', '1/m^2/s', 'rt'):
            return data['DIFF']['CHI_I'].data()

        if self.register(key, ['Qconde', 'electron conductive heat flux'], 'Electron conductive heat flux', 'W/m^2', 'rt'):
            return data['DIFF']['QCONDE'].data()

        if self.register(key, ['Qcondi', 'ion conductive heat flux'], 'Ion conductive heat flux', 'W/m^2', 'rt'):
            return data['DIFF']['QCONDI'].data()

        if self.register(key, ['Qconve', 'electron convective heat flux'], 'Electron convective heat flux', 'W/m^2', 'rt'):
            return data['DIFF']['QCONVE'].data()

        if self.register(key, ['Qconvi', 'ion convective heat flux'], 'Ion convective heat flux', 'W/m^2', 'rt'):
            return data['DIFF']['QCONVI'].data()

        if self.register(key, ['Qcondx', 'exchange heat source'], 'Exchange heat source', 'J/m^3', 'rt'):
            return data['HCD']['Q_EI'].data()

        if self.register(key, ['grho', '<grad*rho>'], 'surface averaged gradient rho', '', 'rt'):
            return data['EQU']['GRAD1RHO'].data()

        if self.register(key, ['Qe', 'total anomalous electron heat flux'], 'Total anomalous electron heat flux (Qconde + Qconve - Qx)', 'W/m^2', 'rt'):
            return data['DIFF']['E_FLUXE'].data()

        if self.register(key, ['Qi', 'total anomalous ion heat flux'], 'Total anomalous ion heat flux (Qcondi + Qconvi + Qx)', 'W/m^2', 'rt'):
            return data['DIFF']['E_FLUXI'].data()

        if self.register(key, ['Ge', 'total anomalous electron particle flux'], 'Total anomalous electron particle flux', '1/s/m^2', 'rt'):
            return data['DIFF']['P_FLUX_ELCT'].data()

        if self.register(key, ['Gi', 'total anomalous ion particle flux'], 'Total anomalous ion particle flux', '1/s/m^2', 'rt'):
            return data['DIFF']['P_FLUX_ION'].data()

        if self.register(key, ['Pr', 'total anomalous momentum flux'], 'Total anomalous momentum flux', 'kg/s^2', 'rt'):
            return data['DIFF']['ROT_FLUX'].data()

        # -------------------------
        # radial derivatives
        # -------------------------
        for k in list(self['__quantities__'].keys()):
            if k.startswith('d('):
                continue
            for dd in ['r_min', 'rho_m', 'psi', 'rho']:
                if self['__quantities__'][k]['raw_dimension'] in ['rt', 'irt', 'brt']:
                    if self.register(key, ['d(' + k + ')/d(' + dd + ')'], 'd/dr(' + self['__quantities__'][k]['description'] + ')', self['__quantities__'][k]['units'] + '/' + self['__quantities__'][dd]['units'], self['__quantities__'][k]['raw_dimension']):
                        if self['__quantities__'][k]['raw_dimension'] == 'rt':
                            n = 1
                        else:
                            n = len(self[k])
                        out = []
                        for i in range(n):
                            if self['__quantities__'][k]['raw_dimension'] == 'rt':
                                tmp = self[k]
                            else:
                                tmp = self[k][i]
                            if self['__quantities__'][dd]['raw_dimension'] == 'r':
                                out.append(gradient(tmp, axis=0) / gradient(self[dd], axis=0)[:, np.newaxis])
                            else:
                                out.append(gradient(tmp, axis=0) / gradient(self[dd], axis=0))
                        if self['__quantities__'][k]['raw_dimension'] == 'rt':
                            return out[0]
                        else:
                            return np.array(out)


######################################################
class ONETWO_statefiles_adaptor(OMFITinterfaceAdaptor):
    type = 'ONETWO_statefiles' # this adapts the data output by ONETWOtime

    @staticmethod
    def identify(data):

        from omfit_classes.omfit_base import OMFITcollection
        printd('Attempting to identify {:} type...'.format('ONETWO_statefiles'))
        if isinstance(data, OMFITcollection) and np.all([k in data for k in ['press', 'rho_grid']]):
            printd(' > Looks like it is {:} type!'.format('ONETWO_statefiles'))
            return True
        else:
            printd(' > It does not seem to be {:} type.'.format('ONETWO_statefiles'))

    def gather_timeslices(self, key, data):
        import numpy as np
        import numpy
        printd('gather timeslices...')
        times = list(data.keys())
        printd(' > gather {:}... found timeslices: {:}'.format(key, times))
        sh = np.shape(data[times[0]][key]['data'])
        if len(sh) == 0:  # time dependent only
            nx = 0
            ns = 0
            var = np.zeros(len(times))
        else:
            nx = sh[-1]  # length of position grid
            if len(sh) > 1:
                ns = sh[0]  # number of species
            else:
                ns = 0  # 0 means the quantity doesn't vary by species
            if ns > 0:
                var = np.zeros([ns, nx, len(times)])
            else:
                var = np.zeros([nx, len(times)])
        printd(' >> expected final shape = ', np.shape(var))
        for i, t in enumerate(times):
            newslice = data[t][key]['data']
            # printd(' >> np.shape(newslice)=',np.shape(newslice))
            if nx == 0:
                var[i] = newslice
            else:
                if ns == 0:
                    var[:, i] = newslice
                else:
                    var[:, :, i] = newslice
        printd(' >> np.shape(var)={:}'.format(np.shape(var)))
        return var

    def define_from_first_slice(self, key, data):
        printd('define from first slice...')
        t = list(data.keys())[0]
        printd(' > using first slice ({:} ms) to define {:}'.format(t, key))
        var = data[t][key]  # some applications will require you to take var['data'] later
        return var

    def definitions(self, key, data):
        import numpy as np
        import numpy
        printd('ONETWO_statefiles definitions for key {:}...'.format(key))

        def interp_npsi_vars_to_rho(val):
            printd(' > interp_npsi_vars_to_rho')
            rho = self.gather_timeslices('rho_grid', data)
            rho_mhd = self.gather_timeslices('rho_mhd_gridnpsi', data)  # rho_mhd and val had both been modified using flipud. I think this just reverses 1d arrays
            printd('  >> np.shape(val) = {:}'.format(np.shape(val)))
            # val = flipud(val)
            printd('  >> np.shape(val) = {:}'.format(np.shape(val)))
            printd('  >> np.shape(rho_mhd) = {:}'.format(np.shape(rho_mhd)))
            printd('  >> np.shape(rho) = {:}'.format(np.shape(rho)))
            nt = len(data)
            out = np.zeros([len(rho[:, 0]), nt])
            for i in range(nt):
                out[:, i] = interp1e(rho_mhd[:, i], val[:, i], kind='linear')(rho[:, i])
            printd('  >>> np.shape(out) = {:}'.format(np.shape(out)))
            return out

        if self.register(key, 'time', 'Time', 's', 't'):
            printd(' > returning time data')
            result = self.gather_timeslices('time', data)
            printd(' >> result = {:}'.format(result))
            return result

        # -------------------------
        # Geometry
        # -------------------------
        if self.register(key, 'rho', 'Normalized minor radius', '1', 'r'):
            printd(' > rho')
            rhogrid = self.define_from_first_slice('rho_grid', data)['data']
            nx = len(rhogrid)
            printd('nx=', nx)
            printd('np.shape(rhogrid)=', np.shape(rhogrid), 'rhogrid.min()=', rhogrid.min(), 'rhogrid.max()=', rhogrid.max())
            result = np.linspace(0, 1, nx)
            printd('rho: np.shape(result) =', np.shape(result), 'result.min()=', result.min(), 'result.max()=', result.max())
            return result

        if self.register(key, 'rho_m', 'Minor radius', 'm', 'rt'):
            return self.gather_timeslices('rho_grid', data)

        if self.register(key, 'a', 'Plasma midplane minor radius', 'm', 't'):
            result = interp_npsi_vars_to_rho(self.gather_timeslices('rminavnpsi', data))[-1, :]  # -1 to select the end of the position array
            printd(' >> result = {:}'.format(result))
            return result

        if self.register(key, 'psi', 'Poloidal flux', 'V*s/rad', 'rt'):
            return self.gather_timeslices('psir_grid', data)

        if self.register(key, ['psin', 'psi_n'], 'Normalized poloidal flux', '1', 'rt'):
            posaxis = 0
            return (self['psi'] - self['psi'].min(axis=posaxis)[np.newaxis, :]) / ((self['psi'].max(axis=posaxis) - self['psi'].min(axis=posaxis))[np.newaxis, :])

        if self.register(key, 'r_min', 'Geometric midplane minor radius', 'm', 'rt'):
            return interp_npsi_vars_to_rho(self.gather_timeslices('rminavnpsi', data))

        if self.register(key, ['r_geo', 'r_maj'], 'Geometric midplane major radius', 'm', 'rt'):
            printd('r_geo')
            result = interp_npsi_vars_to_rho(self.gather_timeslices('rmajavnpsi', data))
            printd('np.shape(result)= {:}'.format(np.shape(result)))
            return result

        if self.register(key, ['rout', 'r_out'], 'Geometric midplane outer major radius', 'm', 'rt'):
            return interp_npsi_vars_to_rho(self.gather_timeslices('rmajavnpsi', data) +
                                           self.gather_timeslices('rminavnpsi', data))  # does this make sense?

        if self.register(key, ['R0', 'vacuum major radius'], 'Major radius of vacuum BT0 reference == R0', 'm', ''):
            return self.gather_timeslices('rmajor', data)[0]

        if self.register(key, '1/R', '<1/R>', '1/m', 'rt'):
            return interp_npsi_vars_to_rho(self.gather_timeslices('ravginpsi', data))

        if self.register(key, 'kappa', 'Elongation', '1', 'rt'):
            return interp_npsi_vars_to_rho(self.gather_timeslices('elongxnpsi', data))

        if self.register(key, 'delta', 'Triangularity', '1', 'rt'):
            return interp_npsi_vars_to_rho(0.5 * self.gather_timeslices('triangnpsi_l', data) + 0.5 * self.gather_timeslices('triangnpsi_u', data))

        if self.register(key, ['Acx', 'plasma cross section'], 'Total plasma cross section', 'm^2', 't'):
            return self.gather_timeslices('areao', data)

        if self.register(key, 'area', 'Cross sectional area enclosed by each flux surface', 'm^2', 'rt'):
            return self.gather_timeslices('cxareanrho', data)

        if self.register(key, 'darea', 'Gradient of cross sectional area enclosed by each flux surface', 'm^2', 'rt'):
            return gradient(self['area'])[0]

        if self.register(key, ['vol', 'V', 'volume'], 'Volume', 'm^3', 'rt'):
            printd('volume')
            integrand = 4 * np.pi ** 2 * self['R0'] * self['rho_m'] * self.gather_timeslices('hcap', data)
            positionaxis = 0
            return scipy.integrate.cumtrapz(integrand, self['rho_m'], initial=0, axis=positionaxis)

        if self.register(key, ['dvol', 'dV/drho', 'differential volume'], 'Differential volume', 'm^3', 'rt'):
            positionaxis = 0  # I know I could just put the 0 directly on the next line, but I want to be very clear about which axis I want
            return gradient(self['volume'])[positionaxis] / gradient(self['rho'])[positionaxis]

        if self.register(key, 'pprime', 'dp/dpsi', 'N/m^2', 'rt'):
            positionaxis = 0
            return gradient(self['pressure'])[positionaxis] / gradient(self['psi'])[positionaxis]

        # -------------------------
        # Fields
        # -------------------------
        if self.register(key, ['BT0', 'vacuum toroidal field'], 'Vacuum BT0 at R0', 'T', 't'):
            return self.gather_timeslices('btor', data)

        if self.register(key, ['f', 'fpol', 'FPOL', 'I_psi'], 'Bz * R (including dia/paramagnetism)', 'm*T', 'rt'):  # f of psi = R * Bt
            return self.gather_timeslices('fpsinrho', data)

        if self.register(key, 'ffprime', 'f*df/dpsi', '(m*T)^2/(V*s/rad)', 'rt'):
            return self.gather_timeslices('ffprim', data)

        if self.register(key, ['Btot_out', 'total B field out', 'B magnitude out', 'btot_out'], 'B field magnitude at outboard midplane', 'T', 'rt'):
            return self.gather_timeslices('btotrmaj', data)

        if self.register(key, ['Bp_avg', 'Bp_fsa', 'BPOL_AVG', 'Bpol_fsa'], 'Flux surface average of poloidal B field', 'T', 'rt'):
            return self.gather_timeslices('bp', data)

        # -------------------------
        # Current
        # -------------------------
        if self.register(key, ['Ip', 'total current'], 'Total plasma current ', 'A', 't'):
            return self.gather_timeslices('tot_cur', data)

        if self.register(key, ['jtot', 'current'], 'Current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.gather_timeslices('curden', data)

        if self.register(key, ['jpar', 'j//', 'parallel current'], 'Parallel current density <J.B/Bt0>', 'A/m^2', 'rt'):
            return self.gather_timeslices('curpar', data)

        if self.register(key, ['joh', 'johm', 'ohmic current'], 'Ohmic current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.gather_timeslices('curohm', data)

        if self.register(key, ['jboot', 'bootstrap current'], 'Bootstrap current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.gather_timeslices('curboot', data)

        if self.register(key, ['jb', 'beam current'], 'Beam current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.gather_timeslices('curbeam', data)

        if self.register(key, ['jrf', 'rf current'], 'RF current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.gather_timeslices('currf', data)

        if self.register(key, ['q_abs', 'safety factor magnitude'], 'Safety factor magnitude', '', 'rt'):
            return abs(self.gather_timeslices('q_value', data))  ###

        # -------------------------
        # Energy, pressure and power
        # -------------------------
        if self.register(key, ['betat'], 'Toroidal beta', '', 't'):
            return self.gather_timeslices('beta', data)

        if self.register(key, ['betap'], 'Poloidal beta', '', 't'):
            return self.gather_timeslices('betap', data)

        if self.register(key, ['betan'], 'Normalized beta', '', 't'):
            bt = self['BT0'] * self['R0'] / self['r_geo'][-1, :]
            return abs(self['betat'] / (self['Ip'] / 1E6 / self['a'] / bt))

        if self.register(key, ['Wtot', 'energy'], 'Total energy', 'J/m^3', 'rt'):
            return self['pressure'] * 3 / 2.

        if self.register(key, ['ptot', 'pressure'], 'Total pressure', 'N/m^2', 'rt'):
            return self.gather_timeslices('press', data)

        if self.register(key, ['Wbeam', 'beam energy'], 'Beam energy density', 'J/m^3', 'rt'):
            return self.gather_timeslices('wbeam', data) * 1.60218e-16  # Conversion from keV/m^3 to J/m^3=N/m^2

        if self.register(key, ['pbeam', 'beam pressure'], 'Beam pressure', 'N/m^2', 'rt'):
            return self['beam energy'] * 2 / 3.

        # -------------------------
        # Temperature and densities
        # -------------------------
        if self.register(key, ['Te', 'electron temperature'], 'Electron temperature', 'keV', 'rt'):
            return self.gather_timeslices('Te', data)

        if self.register(key, ['ne', 'electron density'], 'Electron density', '1/m^3', 'rt'):
            return self.gather_timeslices('ene', data)

        if self.register(key, ['Ti', 'ion temperature'], 'Ion temperature', 'keV', 'irt'):
            return np.array([self.gather_timeslices('Ti', data)] * len(self['ion species']))

        if self.register(key, 'ion species', 'Ion species', '1', 'i'):
            result = self.define_from_first_slice('enion', data)['long_name'].split(':')[1].strip().split()
            printd(' >> result =', result)
            return result  # both primary and impurity ions

        if self.register(key, ['ni', 'ion densities'], 'Ion densities', '1/m^3', 'irt'):
            result = self.gather_timeslices('enion', data)
            printd('  >> np.shape(result) = ', np.shape(result))
            return result  # includes both primary and impurity ions

        if self.register(key, 'beam species', 'Beam ion species', '1', 'b'):
            printd(' > beam species')
            longname = self.define_from_first_slice('enbeam', data)['long_name']
            result = longname.split(':')[1].strip().split()
            printd(' longname =', longname)
            printd(' result = ', result)
            return result

        if self.register(key, ['nb', 'beam densities'], 'Beam ion densities', '1/m^3', 'brt'):
            printd(' > beam densities')
            result = self.gather_timeslices('enbeam', data)
            printd('  >> np.shape(result) = ', np.shape(result))
            return result  # includes all beam species

        # -------------------------
        # Other plasma parameters
        # -------------------------
        if self.register(key, 'Zeff', 'Effective impurity charge', 'amu', 'rt'):
            result = self.gather_timeslices('zeff', data)
            printd('  >> zeff: np.shape(result) = ', np.shape(result))
            return result

        if self.register(key, ['Z', 'Zi', 'charge state'], 'Ion charge state', 'amu', 'i'):
            result = self.define_from_first_slice('z', data)['data'][:, 0]
            printd('  >> zi: np.shape(result) = ', np.shape(result))
            return result

        if self.register(key, ['ftrap', 'trapfrac', 'trapped fraction', 'ncft'], 'Trapped particle fraction', '', 'rt'):
            result = self.gather_timeslices('ftrap', data)
            printd('  >> ftrap: np.shape(result) = ', np.shape(result))
            return result

        if self.register(key, ['electron collisionality', 'nuste', 'nuestar'], 'Electron collisionality', '', 'rt'):
            return self.gather_timeslices('xnuse', data)

        if self.register(key, ['ion collisionality', 'nusti', 'nuistar'], 'Ion collisionality', '', 'irt'):
            return self.gather_timeslices('xnus', data)

        if self.register(key, ['resist'], 'Sauter neoclassical resistivity', 'Ohm m', 'rt'):
            return self.gather_timeslices('eta', data)


######################################################
class TRANSP_MDS_adaptor(OMFITinterfaceAdaptor):
    type = 'TRANSP_MDS'

    @staticmethod
    def identify(data):
        printd('Attempting to identify {:} type...'.format('TRANSP_MDS'))
        if 'OUTPUTS' in data and data['OUTPUTS']['ONE_D']['Q0'].data() is not None:
            printd(' > Looks like it is {:} type!'.format('TRANSP_MDS'))
            return True
        else:
            printd(' > It does not seem to be {:} type.'.format('TRANSP_MDS'))

    def tr2D(self, key, data):
        import numpy as np
        import numpy
        # * X is the normalized radius of the zone centers
        # * XB is the normalized radius of the zone boundaries
        # * RMAJM is the outer midplane radius defined at XB
        # TRANSP grids are not defined on axis, hence use linear extrapolation
        xname = data['OUTPUTS']['TWO_D'][key]['XAXIS'].data()[0].strip()
        y = data['OUTPUTS']['TWO_D'][key].data()
        if y is None:
            return np.zeros(self['rho'].size, self['time'].size)
        y = y.T
        # internally we convert all TRANSP outputs to normalized radius of the zone boundaries: XB
        xb = data['OUTPUTS']['TWO_D']['XB'].data().T[:, 0]
        xb0 = np.hstack((0, xb))
        if xname in ['X', 'XB']:
            x = data['OUTPUTS']['TWO_D'][xname].data().T[:, 0]
            tmp = interp1e(x, y.T)(xb0).T
            return tmp
        else:
            raise Exception(key + ': ' + xname + ' axis interpolation to XB has not been coded up yet!')

    def definitions(self, key, data):
        import numpy as np
        import numpy
        printd('TRANSP_MDS definitions...')
        if self.register(key, 'device', 'Device', '1', ''):
            return data['DEVICE'].data()

        if self.register(key, 'shot', 'Shot', '1', ''):
            return int(data['SOURCE_SHOT'].data())

        if self.register(key, 'time', 'Time', 's', 't'):
            return data['TRANSP_OUT']['TIME2D'].data()

        # -------------------------
        # Geometry
        # -------------------------
        if self.register(key, ['R0', 'vacuum major radius'], 'Major radius of vacuum BT0 reference == R0', 'm', ''):
            specs = omfit_classes.utils_fusion.device_specs(tolist(self['device'])[0])
            if 'R0' in list(specs.keys()) and specs['R0'] is not None:
                return specs['R0']
            else:
                raise Exception('Plasma vacuum major radius is not defined for ' + str(self['device']))

        if self.register(key, 'rho_m', 'Minor radius', 'm', 'rt'):
            return self.tr2D('RBOUN', data) / 100.

        if self.register(key, 'a', 'Plasma midplane minor radius', 'm', 't'):
            return self['r_min'][-1, :]

        if self.register(key, 'r_min', 'Geometric midplane minor radius', 'm', 'rt'):
            return self.tr2D('RMNMP', data) / 100.

        if self.register(key, ['r_geo', 'r_maj'], 'Geometric midplane major radius', 'm', 'rt'):
            return self.tr2D('RMJMP', data) / 100.

        if self.register(key, ['rout', 'r_out'], 'Geometric midplane outer major radius', 'm', 'rt'):
            return self['r_geo'] + self['r_min']

        if self.register(key, 'rho', 'Normalized minor radius', '1', 'r'):
            return self.tr2D('XB', data)[:, 0]

        if self.register(key, '1/R', '<1/R>', '1/m', 'rt'):
            return self.tr2D('GRI', data) * 100

        if self.register(key, ['eps', 'inverse aspect ratio'], 'epsilon: Inverse aspect ratio', '', 'rt'):
            return 1.0 / self.tr2D('ARAT', data)

        if self.register(key, 'kappa', 'Elongation', '1', 'rt'):
            return self.tr2D('ELONG', data)

        if self.register(key, 'delta', 'Triangularity', '1', 'rt'):
            return self.tr2D('TRIANG', data)

        if self.register(key, ['Acx', 'plasma cross section'], 'Total plasma cross section', 'm^2', 't'):
            return data['OUTPUTS']['ONE_D']['PAREA'].data() / 1E4  # This might actually be cross section of psi_N=0.995

        if self.register(key, ['darea'], 'Gradient of cross sectional area enclosed by each flux surface', 'm^2', 'rt'):
            return self.tr2D('DAREA', data) / 1e4

        if self.register(key, ['area'], 'Cross sectional area enclosed by each flux surface', 'm^2', 'rt'):
            return np.cumsum(self['darea'], axis=0)  # Inconsistent w/ Acx because of interpolation; summing raw DAREA out
            #                                      to the edge agrees perfectly with PAREA, but the position baxis for
            #                                      DAREA is X, which ends at 0.995. Extrapolating to 1 makes the result
            #                                      larger.

        if self.register(key, ['vol', 'V', 'volume'], 'Volume', 'm^3', 'rt'):
            rho_rt = np.tile(self['rho'], (self['dvol'].shape[1], 1)).T
            return scipy.integrate.cumtrapz(self['dvol'], rho_rt, axis=0, initial=0)

        if self.register(key, ['dvol', 'dV/drho', 'differential volume'], 'Differential volume', 'm^3', 'rt'):
            return self.tr2D('DVOL', data) / gradient(self['rho'])[:, np.newaxis] / 1E6

        if self.register(key, 'psi', 'Poloidal flux', 'V*s/rad', 'rt'):
            return self.tr2D('PLFLX', data)

        if self.register(key, ['psin', 'psi_n'], 'Normalized poloidal flux', '1', 'rt'):
            return (self['psi'] - np.min(self['psi'], 0)[np.newaxis, :]) / (np.max(self['psi'], 0)[np.newaxis, :] - np.min(self['psi'], 0)[np.newaxis, :])

        if self.register(key, 'pprime', 'dp/dpsi', 'N/m^2', 'rt'):
            return gradient(self['pressure'])[0] / gradient(self['psi'])[0]

        # -------------------------
        # Fields
        # -------------------------
        if self.register(key, ['BT0', 'vacuum toroidal field'], 'Vacuum BT0 at R0', 'T', 't'):
            return data['OUTPUTS']['ONE_D']['BZXR'].data() / (self['R0'] * 100)

        if self.register(key, ['f', 'fpol', 'FPOL', 'I_psi'], 'Bz * R (including dia/paramagnetism)', 'm*T', 'rt'):
            return self.tr2D('GFUN', data) * data['OUTPUTS']['ONE_D']['BZXR'].data()[np.newaxis, :] / 1E2

        if self.register(key, 'fprime', 'df/dpsi', '(m*T)/(V*s/rad)', 'rt'):
            return gradient(self['f'])[0] / gradient(self['psi'])[0]

        if self.register(key, 'ffprime', 'f*df/dpsi', '(m*T)^2/(V*s/rad)', 'rt'):
            return self['f'] * self['fprime']

        if self.register(key, ['<B>', 'GB1', 'B fsa'], '<B>', 'T', 'rt'):
            return self.tr2D('GB1', data)

        if self.register(key, ['<B**2>', 'GB2', 'B sq fsa'], '<B^2>', 'T^2', 'rt'):
            return self.tr2D('GB2', data)

        if self.register(key, ['<B**-2>', 'GB2I', 'inv B sq fsa'], '<B^-2>', 'T^-2', 'rt'):
            return self.tr2D('GB2I', data)

        # -------------------------
        # Current
        # -------------------------
        if self.register(key, ['Ip', 'total current'], 'Total plasma current ', 'A', 't'):
            return data['OUTPUTS']['ONE_D']['PCURC'].data()

        if self.register(key, ['jtot', 'current'], 'Current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('CUR', data) * (1E2) ** 2

        if self.register(key, ['joh', 'johm', 'ohmic current'], 'Ohmic current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('CUROH', data) * (1E2) ** 2

        if self.register(key, ['jboot', 'bootstrap current'], 'Bootstrap current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('CURBS', data) * (1E2) ** 2

        if self.register(key, ['j.B fsa', 'jboot.B', 'bootstrap current times B fsa'], 'Flux surf avg of B . neoclassical bootstrap current density <Jtor R0/R>', 'T*A/m^2', 'rt'):
            return self.tr2D('PLJBSNC', data) * (1E2) ** 2

        if self.register(key, ['jbeam', 'beam current'], 'Beam current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('CURB', data) * (1E2) ** 2

        if self.register(key, ['jec', 'ec current'], 'EC current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('ECCUR', data) * (1E2) ** 2

        if self.register(key, ['jlh', 'lh current'], 'LH current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('LHCUR', data) * (1E2) ** 2

        if self.register(key, ['jrf', 'rf current'], 'RF current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self['ec current'] + self['rf current']

        if self.register(key, ['q_abs', 'safety factor magnitude'], 'Safety factor magnitude', '', 'rt'):
            return self.tr2D('Q', data)

        # -------------------------
        # Energy, pressure and power
        # -------------------------
        if self.register(key, ['betat'], 'Toroidal beta', '', 't'):
            dvol = data['OUTPUTS']['TWO_D']['DVOL'].data()
            bttot = data['OUTPUTS']['TWO_D']['BTTOT'].data()
            return np.sum(bttot * dvol, 1) / np.sum(dvol, 1)

        # if self.register(key,['betap'],'Poloidal beta','','t'):
        #     return

        if self.register(key, ['betan'], 'Normalized beta', '', 't'):
            bt = self['BT0'] * self['R0'] / self['r_geo'][-1, :]
            return abs(self['betat'] / (self['Ip'] / 1E6 / self['a'] / bt))

        if self.register(key, ['Wtot', 'energy'], 'Total energy', 'J/m^3', 'rt'):
            # return self.tr2D('UTOTL',data)*(1E2)**3
            return self['pressure'] * 3 / 2.

        if self.register(key, ['ptot', 'pressure'], 'Total pressure', 'N/m^2', 'rt'):
            # return self['energy']*2/3.
            return self.tr2D('PTOWB', data)

        if self.register(key, ['Wth', 'thermal energy'], 'Thermal energy', 'J/m^3', 'rt'):
            return self.tr2D('UTHRM', data) * (1E2) ** 3

        if self.register(key, ['Pth', 'thermal pressure'], 'Thermal pressure', 'N/m^2', 'rt'):
            return self['thermal energy'] * 2 / 3.

        if self.register(key, ['Wbeam', 'beam energy'], 'Beam energy density', 'J/m^3', 'rt'):
            return self['energy'] - self['thermal energy']

        if self.register(key, ['pbeam', 'beam pressure'], 'Beam pressure', 'N/m^2', 'rt'):
            return self['pressure'] - self['thermal pressure']

        # -------------------------
        # Temperature and densities
        # -------------------------
        if self.register(key, ['Te', 'electron temperature'], 'Electron temperature', 'keV', 'rt'):
            return self.tr2D('TE', data) / 1E3

        if self.register(key, ['ne', 'electron density'], 'Electron density', '1/m^3', 'rt'):
            return self.tr2D('NE', data) * 1E6

        if self.register(key, ['<ne>'], 'Volume averaged electron density', '', 't'):
            dvol = data['OUTPUTS']['TWO_D']['DVOL'].data()
            ne = data['OUTPUTS']['TWO_D']['NE'].data() * 1E6
            return np.sum(ne * dvol, 1) / np.sum(dvol, 1)

        if self.register(key, ['fGW'], 'Greenwald fraction', '', 't'):
            nGW = (self['Ip'] / 1E6 / np.pi / self['a'] ** 2)
            return self['<ne>'] / 1E20 / nGW

        if self.register(key, 'ion species', 'Ion species', '1', 'i'):
            species = []
            for s in ['3', '4', '6', 'H', 'D', 'T', 'M']:
                if 'N' + s in data['OUTPUTS']['TWO_D']:
                    species.append(s)
            # First search for individually listed specific impurities named NIMPS_*
            imp_found = False
            for s in ['C', 'Ne', ]:  # List of impurity species to search for
                if 'NIMPS_{:}'.format(s) in data['OUTPUTS']['TWO_D']:
                    species.append(s)
                    imp_found = True
            # If we can't find specific impurities listed individually, try the generic NIMP and guess which impurity
            # it is based on which machine.
            if imp_found:
                printd('Found specifically listed impurities in TRANSP statefile.')
            else:
                if 'NIMP' in data['OUTPUTS']['TWO_D']:
                    if self['device'] == 'D3D':
                        s = 'C'
                    else:
                        # Generic species guess for unknown tokamak
                        s = 'C'
                    species.append(s)
                    printd('Could not find specifically listed impurities in TRANSP statefile so used NIMP.')
            return species

        if self.register(key, ['Ti', 'ion temperature'], 'Ion temperature', 'keV', 'irt'):
            return np.array([self.tr2D('TI', data)] * len(self['ion species'])) / 1E3

        if self.register(key, ['ni', 'ion densities'], 'Ion densities', '1/m^3', 'irt'):
            nion = []
            for s in ['3', '4', '6', 'H', 'D', 'T', 'M']:
                if 'N' + s in data['OUTPUTS']['TWO_D']:
                    nion.append(self.tr2D('N' + s, data) * 1E6)
            # First search for individually listed specific impurities named NIMPS_*
            imp_found = False
            for s in ['C', 'Ne', ]:
                if 'NIMPS_{:}'.format(s) in data['OUTPUTS']['TWO_D']:
                    imp_found = True
                    nion.append(self.tr2D('NIMPS_{:}'.format(s), data) * 1e6)
            # If we can't find specific impurities listed individually, try the generic NIMP and guess which impurity
            # it is based on which machine.
            if imp_found:
                printd('Found specifically listed impurities in TRANSP statefile.')
            else:
                if 'NIMP' in data['OUTPUTS']['TWO_D']:
                    nion.append(self.tr2D('NIMP', data) * 1e6)
                    printd('Could not find specifically listed impurities in TRANSP statefile so used NIMP.')
            return np.array(nion)

        if self.register(key, 'beam species', 'Beam ion species', '1', 'b'):
            species = []
            for s in ['H', 'D', 'T', 'He4', 'He3']:
                if 'NB_F1_' + s in data['OUTPUTS']['TWO_D']:
                    species.append(s)
            return species

        if self.register(key, ['nb', 'beam densities'], 'Beam ion densities', '1/m^3', 'brt'):
            nion = []
            for s in ['H', 'D', 'T', 'He4', 'He3']:
                if 'NB_F1_' + s in data['OUTPUTS']['TWO_D']:
                    tot_s = self.tr2D('NB_F' + str(1) + '_' + s, data) * 1E6
                    for k in range(2, 4):
                        tot_s += self.tr2D('NB_F' + str(k) + '_' + s, data) * 1E6
                    nion.append(tot_s)
            return np.array(nion)

        # -----------------------
        # Other plasma parameters
        # -----------------------
        if self.register(key, 'Zeff', 'Effective impurity charge', 'amu', 'rt'):
            return self.tr2D('ZEFFP', data)

        if self.register(key, ['Z', 'Zi', 'charge state'], 'Ion charge state', 'amu', 'i'):
            # This could also have been implemented by looking for ZIMPS_* and then falling back to XZIMP if not found.
            s = self['ion species']
            z_lookup = {'3': 3, '4': 4, '6': 6,
                        'H': 1, 'D': 1, 'T': 1,
                        'He': 2, 'Li': 3, 'Be': 4, 'B': 5,
                        'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10,
                        'M': 0}  # I don't know what this is
            return [z_lookup.get(ss, 0) for ss in s]  # Defaults to zero if it can't find the element in its table

        if self.register(key, ['ftrap', 'trapfrac', 'trapped fraction', 'ncft'], 'Trapped particle fraction', '', 'rt'):
            result = self.tr2D('NCFT', data)
            printd('  >> ftrap: np.shape(result) = ', np.shape(result))
            return result

        if self.register(key, ['resist'], 'Sauter neoclassical resistivity', 'Ohm m', 'rt'):
            return self.tr2D('ETA_SNC', data) / 100.

        if self.register(key, ['spitzer'], 'Spitzer resistivity', 'Ohm m', 'rt'):
            return self.tr2D('ETA_SP', data) / 100.

        if self.register(key, ['lnLambda_e'], 'Coulomb logarithm lnLambda for electrons', '', 'rt'):
            return self.tr2D('CLOGE', data)

        if self.register(key, ['lnLambda_i'], 'Coulomb logarithm lnLambda for ions', '', 'rt'):
            return self.tr2D('CLOGI', data)

        if self.register(key, ['electron collisionality', 'nuste', 'nuestar'], 'Electron collisionality', '', 'rt'):
            return self.tr2D('NUSTE', data)

        if self.register(key, ['ion collisionality', 'nusti', 'nuistar'], 'Ion collisionality', '', 'irt'):
            return np.array([self.tr2D('NUSTI', data)] * len(self['ion species']))


######################################################
class TRANSP_CDF_adaptor(OMFITinterfaceAdaptor):
    type = 'TRANSP_CDF'

    @staticmethod
    def identify(data):
        printd('Attempting to identify {:} type...'.format('TRANSP_CDF'))
        if 'NZONES' in data.get('_globals', {}):
            printd(' > Looks like it is {:} type!'.format('TRANSP_CDF'))
            return True
        else:
            printd(' > It does not seem to be {:} type.'.format('TRANSP_CDF'))

    def tr2D(self, key, data):
        """
        Interpolates all 2D data from transp onto a consistent "zone boundary" grid,
        including interpolation to axis.

        :param key: str.
        :param data: OMFITnc. netcdf TRANSP output
        :return:

        """
        import numpy as np
        import numpy
        # * X is the normalized radius of the zone centers
        # * XB is the normalized radius of the zone boundaries
        # * RMAJM is the outer midplane radius defined at XB
        # TRANSP grids are not defined on axis, hence use linear extrapolation
        xname = data[key]['__dimensions__'][1].strip()
        y = data[key]['data']
        if y is None:
            return np.zeros(self['rho'].size, self['time'].size)
        y = y.T
        # internally we convert all TRANSP outputs to normalized radius of the zone boundaries: XB
        xb = data['XB']['data'].T[:, 0]
        xb0 = np.hstack((0, xb))
        if xname in ['X', 'XB']:
            x = data[xname]['data'].T[:, 0]
            tmp = interp1e(x, y.T)(xb0).T
            return tmp
        else:
            raise Exception(key + ': ' + xname + ' axis interpolation to XB has not been coded up yet!')

    def definitions(self, key, data):
        import numpy as np
        import numpy
        printd('TRANSP_CDF definitions...')
        if self.register(key, 'device', 'Device', '1', ''):
            return tokamak(data['_globals'].get('device', 'UNKNOWN'))

        if self.register(key, 'shot', 'Shot', '1', ''):
            return int(data['_globals']['shot'])

        if self.register(key, 'time', 'Time', 's', 't'):
            return data['TIME']['data']

        # -------------------------
        # Geometry
        # -------------------------
        if self.register(key, ['R0', 'vacuum major radius'], 'Major radius of vacuum BT0 reference == R0', 'm', ''):
            specs = omfit_classes.utils_fusion.device_specs(tolist(self['device'])[0])
            if 'R0' in specs.keys() and specs['R0'] is not None:
                return specs['R0']
            else:
                raise Exception('Plasma vacuum major radius is not defined for ' + str(self['device']))

        if self.register(key, 'rho_m', 'Minor radius', 'm', 'rt'):
            return self.tr2D('RBOUN', data) / 100.

        if self.register(key, 'a', 'Plasma midplane minor radius', 'm', 't'):
            return self['r_min'][-1, :]

        if self.register(key, 'r_min', 'Geometric midplane minor radius', 'm', 'rt'):
            return self.tr2D('RMNMP', data) / 100.

        if self.register(key, ['r_geo', 'r_maj'], 'Geometric midplane major radius', 'm', 'rt'):
            return self.tr2D('RMJMP', data) / 100.

        if self.register(key, ['rout', 'r_out'], 'Geometric midplane outer major radius', 'm', 'rt'):
            return self['r_geo'] + self['r_min']

        if self.register(key, 'rho', 'Normalized minor radius', '1', 'r'):
            return self.tr2D('XB', data)[:, 0]

        if self.register(key, '1/R', '<1/R>', '1/m', 'rt'):
            return self.tr2D('GRI', data) * 100

        if self.register(key, ['eps', 'inverse aspect ratio'], 'epsilon: Inverse aspect ratio', '', 'rt'):
            return 1.0 / self.tr2D('ARAT', data)

        if self.register(key, 'kappa', 'Elongation', '1', 'rt'):
            return self.tr2D('ELONG', data)

        if self.register(key, 'delta', 'Triangularity', '1', 'rt'):
            return self.tr2D('TRIANG', data)

        if self.register(key, ['Acx', 'plasma cross section'], 'Total plasma cross section', 'm^2', 't'):
            return data['PAREA']['data'] / 1E4  # This might actually be cross section of psi_N=0.995

        if self.register(key, ['darea'], 'Gradient of cross sectional area enclosed by each flux surface', 'm^2', 'rt'):
            return self.tr2D('DAREA', data) / 1e4

        if self.register(key, ['area'], 'Cross sectional area enclosed by each flux surface', 'm^2', 'rt'):
            return np.cumsum(self['darea'], axis=0)  # Inconsistent w/ Acx because of interpolation; summing raw DAREA out
            #                                      to the edge agrees perfectly with PAREA, but the position baxis for
            #                                      DAREA is X, which ends at 0.995. Extrapolating to 1 makes the result
            #                                      larger.

        if self.register(key, ['vol', 'V', 'volume'], 'Volume', 'm^3', 'rt'):
            rho_rt = np.tile(self['rho'], (self['dvol'].shape[1], 1)).T
            return scipy.integrate.cumtrapz(self['dvol'], rho_rt, axis=0, initial=0)

        if self.register(key, ['dvol', 'dV/drho', 'differential volume'], 'Differential volume', 'm^3', 'rt'):
            return self.tr2D('DVOL', data) / gradient(self['rho'])[:, np.newaxis] / 1E6

        if self.register(key, 'psi', 'Poloidal flux', 'V*s/rad', 'rt'):
            return self.tr2D('PLFLX', data)

        if self.register(key, ['psin', 'psi_n'], 'Normalized poloidal flux', '1', 'rt'):
            return (self['psi'] - np.min(self['psi'], 0)[np.newaxis, :]) / (
                    np.max(self['psi'], 0)[np.newaxis, :] - np.min(self['psi'], 0)[np.newaxis, :])

        if self.register(key, 'pprime', 'dp/dpsi', 'N/m^2', 'rt'):
            return gradient(self['pressure'])[0] / gradient(self['psi'])[0]

        # -------------------------
        # Fields
        # -------------------------
        if self.register(key, ['BT0', 'vacuum toroidal field'], 'Vacuum BT0 at R0', 'T', 't'):
            return data['BZXR']['data'] / (self['R0'] * 100)

        if self.register(key, ['f', 'fpol', 'FPOL', 'I_psi'], 'Bz * R (including dia/paramagnetism)', 'm*T', 'rt'):
            return self.tr2D('GFUN', data) * data['BZXR']['data'][np.newaxis, :] / 1E2

        if self.register(key, 'fprime', 'df/dpsi', '(m*T)/(V*s/rad)', 'rt'):
            return gradient(self['f'])[0] / gradient(self['psi'])[0]

        if self.register(key, 'ffprime', 'f*df/dpsi', '(m*T)^2/(V*s/rad)', 'rt'):
            return self['f'] * self['fprime']

        if self.register(key, ['<B>', 'GB1', 'B fsa'], '<B>', 'T', 'rt'):
            return self.tr2D('GB1', data)

        if self.register(key, ['<B**2>', 'GB2', 'B sq fsa'], '<B^2>', 'T^2', 'rt'):
            return self.tr2D('GB2', data)

        if self.register(key, ['<B**-2>', 'GB2I', 'inv B sq fsa'], '<B^-2>', 'T^-2', 'rt'):
            return self.tr2D('GB2I', data)

        # -------------------------
        # Current
        # -------------------------
        if self.register(key, ['Ip', 'total current'], 'Total plasma current ', 'A', 't'):
            return data['PCURC']['data']

        if self.register(key, ['jtot', 'current'], 'Current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('CUR', data) * (1E2) ** 2

        if self.register(key, ['joh', 'johm', 'ohmic current'], 'Ohmic current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('CUROH', data) * (1E2) ** 2

        if self.register(key, ['jboot', 'bootstrap current'], 'Bootstrap current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('CURBS', data) * (1E2) ** 2

        if self.register(key, ['j.B fsa', 'jboot.B', 'bootstrap current times B fsa'],
                         'Flux surf avg of B . neoclassical bootstrap current density <Jtor R0/R>', 'T*A/m^2', 'rt'):
            return self.tr2D('PLJBSNC', data) * (1E2) ** 2

        if self.register(key, ['jbeam', 'beam current'], 'Beam current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('CURB', data) * (1E2) ** 2

        if self.register(key, ['jec', 'ec current'], 'EC current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('ECCUR', data) * (1E2) ** 2

        if self.register(key, ['jlh', 'lh current'], 'LH current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self.tr2D('LHCUR', data) * (1E2) ** 2

        if self.register(key, ['jrf', 'rf current'], 'RF current density <Jtor R0/R>', 'A/m^2', 'rt'):
            return self['ec current'] + self['rf current']

        if self.register(key, ['q_abs', 'safety factor magnitude'], 'Safety factor magnitude', '', 'rt'):
            return self.tr2D('Q', data)

        # -------------------------
        # Energy, pressure and power
        # -------------------------
        if self.register(key, ['betat'], 'Toroidal beta', '', 't'):
            dvol = data['DVOL']['data']
            bttot = data['BTTOT']['data']
            return np.sum(bttot * dvol, 1) / np.sum(dvol, 1)

        # if self.register(key,['betap'],'Poloidal beta','','t'):
        #     return

        if self.register(key, ['betan'], 'Normalized beta', '', 't'):
            bt = self['BT0'] * self['R0'] / self['r_geo'][-1, :]
            return abs(self['betat'] / (self['Ip'] / 1E6 / self['a'] / bt))

        if self.register(key, ['Wtot', 'energy'], 'Total energy', 'J/m^3', 'rt'):
            # return self.tr2D('UTOTL',data)*(1E2)**3
            return self['pressure'] * 3 / 2.

        if self.register(key, ['ptot', 'pressure'], 'Total pressure', 'N/m^2', 'rt'):
            # return self['energy']*2/3.
            return self.tr2D('PTOWB', data)

        if self.register(key, ['Wth', 'thermal energy'], 'Thermal energy', 'J/m^3', 'rt'):
            return self.tr2D('UTHRM', data) * (1E2) ** 3

        if self.register(key, ['Pth', 'thermal pressure'], 'Thermal pressure', 'N/m^2', 'rt'):
            return self['thermal energy'] * 2 / 3.

        if self.register(key, ['Wbeam', 'beam energy'], 'Beam energy density', 'J/m^3', 'rt'):
            return self['energy'] - self['thermal energy']

        if self.register(key, ['pbeam', 'beam pressure'], 'Beam pressure', 'N/m^2', 'rt'):
            return self['pressure'] - self['thermal pressure']

        # -------------------------
        # Temperature and densities
        # -------------------------
        if self.register(key, ['Te', 'electron temperature'], 'Electron temperature', 'keV', 'rt'):
            return self.tr2D('TE', data) / 1E3

        if self.register(key, ['ne', 'electron density'], 'Electron density', '1/m^3', 'rt'):
            return self.tr2D('NE', data) * 1E6

        if self.register(key, ['<ne>'], 'Volume averaged electron density', '', 't'):
            dvol = data['DVOL']['data']
            ne = data['NE']['data'] * 1E6
            return np.sum(ne * dvol, 1) / np.sum(dvol, 1)

        if self.register(key, ['fGW'], 'Greenwald fraction', '', 't'):
            nGW = (self['Ip'] / 1E6 / np.pi / self['a'] ** 2)
            return self['<ne>'] / 1E20 / nGW

        if self.register(key, 'ion species', 'Ion species', '1', 'i'):
            species = []
            for s in ['3', '4', '6', 'H', 'D', 'T', 'M']:
                if 'N' + s in data:
                    species.append(s)
            # First search for individually listed specific impurities named NIMPS_*
            imp_found = False
            for s in ['C', 'Ne', ]:  # List of impurity species to search for
                if 'NIMPS_{:}'.format(s) in data:
                    species.append(s)
                    imp_found = True
            # If we can't find specific impurities listed individually, try the generic NIMP and guess which impurity
            # it is based on which machine.
            if imp_found:
                printd('Found specifically listed impurities in TRANSP statefile.')
            else:
                if 'NIMP' in data:
                    if self['device'] == 'D3D':
                        s = 'C'
                    else:
                        # Generic species guess for unknown tokamak
                        s = 'C'
                    species.append(s)
                    printd('Could not find specifically listed impurities in TRANSP statefile so used NIMP.')
            return species

        if self.register(key, ['Ti', 'ion temperature'], 'Ion temperature', 'keV', 'irt'):
            return np.array([self.tr2D('TI', data)] * len(self['ion species'])) / 1E3

        if self.register(key, ['ni', 'ion densities'], 'Ion densities', '1/m^3', 'irt'):
            nion = []
            for s in ['3', '4', '6', 'H', 'D', 'T', 'M']:
                if 'N' + s in data:
                    nion.append(self.tr2D('N' + s, data) * 1E6)
            # First search for individually listed specific impurities named NIMPS_*
            imp_found = False
            for s in ['C', 'Ne', ]:
                if 'NIMPS_{:}'.format(s) in data:
                    imp_found = True
                    nion.append(self.tr2D('NIMPS_{:}'.format(s), data) * 1e6)
            # If we can't find specific impurities listed individually, try the generic NIMP and guess which impurity
            # it is based on which machine.
            if imp_found:
                printd('Found specifically listed impurities in TRANSP statefile.')
            else:
                if 'NIMP' in data:
                    nion.append(self.tr2D('NIMP', data) * 1e6)
                    printd('Could not find specifically listed impurities in TRANSP statefile so used NIMP.')
            return np.array(nion)

        if self.register(key, 'beam species', 'Beam ion species', '1', 'b'):
            species = []
            for s in ['H', 'D', 'T', 'He4', 'He3']:
                if 'NB_F1_' + s in data:
                    species.append(s)
            return species

        if self.register(key, ['nb', 'beam densities'], 'Beam ion densities', '1/m^3', 'brt'):
            nion = []
            for s in ['H', 'D', 'T', 'He4', 'He3']:
                if 'NB_F1_' + s in data:
                    tot_s = self.tr2D('NB_F' + str(1) + '_' + s, data) * 1E6
                    for k in range(2, 4):
                        tot_s += self.tr2D('NB_F' + str(k) + '_' + s, data) * 1E6
                    nion.append(tot_s)
            return np.array(nion)

        # -----------------------
        # Other plasma parameters
        # -----------------------
        if self.register(key, 'Zeff', 'Effective impurity charge', 'amu', 'rt'):
            return self.tr2D('ZEFFP', data)

        if self.register(key, ['Z', 'Zi', 'charge state'], 'Ion charge state', 'amu', 'i'):
            # This could also have been implemented by looking for ZIMPS_* and then falling back to XZIMP if not found.
            s = self['ion species']
            z_lookup = {'3': 3, '4': 4, '6': 6,
                        'H': 1, 'D': 1, 'T': 1,
                        'He': 2, 'Li': 3, 'Be': 4, 'B': 5,
                        'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10,
                        'M': 0}  # I don't know what this is
            return [z_lookup.get(ss, 0) for ss in s]  # Defaults to zero if it can't find the element in its table

        if self.register(key, ['ftrap', 'trapfrac', 'trapped fraction', 'ncft'], 'Trapped particle fraction', '', 'rt'):
            result = self.tr2D('NCFT', data)
            printd('  >> ftrap: np.shape(result) = ', np.shape(result))
            return result

        if self.register(key, ['resist'], 'Sauter neoclassical resistivity', 'Ohm m', 'rt'):
            return self.tr2D('ETA_SNC', data) / 100.

        if self.register(key, ['spitzer'], 'Spitzer resistivity', 'Ohm m', 'rt'):
            return self.tr2D('ETA_SP', data) / 100.

        if self.register(key, ['lnLambda_e'], 'Coulomb logarithm lnLambda for electrons', '', 'rt'):
            return self.tr2D('CLOGE', data)

        if self.register(key, ['lnLambda_i'], 'Coulomb logarithm lnLambda for ions', '', 'rt'):
            return self.tr2D('CLOGI', data)

        if self.register(key, ['electron collisionality', 'nuste', 'nuestar'], 'Electron collisionality', '', 'rt'):
            return self.tr2D('NUSTE', data)

        if self.register(key, ['ion collisionality', 'nusti', 'nuistar'], 'Ion collisionality', '', 'irt'):
            return np.array([self.tr2D('NUSTI', data)] * len(self['ion species']))


######################################################
class pFile_adaptor(OMFITinterfaceAdaptor):
    # NOTE: this adaptor has hardwired 3 ion species: impurities, beam, main
    type = 'OSBORNE_pFile'

    @staticmethod
    def identify(data):
        import numpy as np
        import numpy
        printd('Attempting to identify {:} type...'.format('OSBORNE_pFile'))
        if isinstance(data, dict) and np.all([k in data for k in ['ne', 'N Z A']]):
            if len(data['N Z A']['Z']) != 3:
                return False
            printd(' > Looks like it is {:} type!'.format('OSBORNE_pFile'))
            return True
        else:
            printd(' > It does not seem to be {:} type.'.format('OSBORNE_pFile'))

    def definitions(self, key, data):
        import numpy as np
        import numpy
        printd('%s definitions...' % self.type)

        if self.register(key, 'psin', 'Normalized poloidal flux', '1', 'r'):
            return np.linspace(0, 1, len(data['ne']['data']))

        # -------------------------
        # Energy, pressure and power
        # -------------------------
        if self.register(key, ['ptot', 'pressure'], 'Total pressure', 'N/m^2', 'r'):
            # return self['energy']*2/3.
            return interp1e(data['ptot']['psinorm'], data['ptot']['data'])(self['psin']) * 1E3

        # -------------------------
        # Temperature and densities
        # -------------------------
        if self.register(key, ['Te', 'electron temperature'], 'Electron temperature', 'keV', 'r'):
            return interp1e(data['te']['psinorm'], data['te']['data'])(self['psin'])

        if self.register(key, ['Ti', 'ion temperature'], 'Ion temperature', 'keV', 'ir'):
            return np.array([interp1e(data['ti']['psinorm'], data['ti']['data'])(self['psin'])] * 2)

        if self.register(key, ['ne', 'electron density'], 'Electron density', '1/m^3', 'r'):
            return interp1e(data['ne']['psinorm'], data['ne']['data'])(self['psin']) * 1E20

        if self.register(key, ['ni', 'ion densities'], 'Ion densities', '1/m^3', 'ir'):
            ni = interp1e(data['ni']['psinorm'], data['ni']['data'])(self['psin']) * 1E20
            nimp = interp1e(data['nz1']['psinorm'], data['nz1']['data'])(self['psin']) * 1E20
            return np.array([ni, nimp])

        if self.register(key, ['nb', 'beam densities'], 'Beam ion densities', '1/m^3', 'br'):
            nb = interp1e(data['nb']['psinorm'], data['nb']['data'])(self['psin']) * 1E20
            return np.array([nb])

        if self.register(key, ['nb', 'beam densities'], 'Beam ion densities', '1/m^3', 'br'):
            nb = interp1e(data['nb']['psinorm'], data['nb']['data'])(self['psin']) * 1E20
            return np.array([nb])

        # -------------------------
        # Other plasma parameters
        # -------------------------
        if self.register(key, 'Zeff', 'Effective impurity charge', 'amu', 'r'):
            # np.sum(ni*Z^2)/np.sum(ni*Z)=np.sum(ni*Z^2)/ne
            ni = self['ni']
            nb = self['nb']
            Zi = data['N Z A']['Z']
            return (ni[1] * Zi[0] ** 2 +  # impurity
                    nb[0] * Zi[1] ** 2 +  # beam
                    ni[0] * Zi[2] ** 2  # main
                    ) / (
                           ni[1] * Zi[0] +
                           nb[0] * Zi[1] +
                           ni[0] * Zi[2]
                   )



"""
Tools for interfacing with Google sheets

Compared to SQL tables, Google sheets allow for formatting and other conveniences
that make them easier to update by hand (appropriate for some columns like labels,
comments, etc.), but the API for Google sheets is relatively simple. This file
is meant to provide tools for looking up data by column header instead of column
number/letter and by shot instead of row number. That is, the user should not have
to know which row holds shot 123456 or which column holds density in order to get
the average density for shot 123456.

In this script, the purely numeric way to refer to cell 'A1' is (0, 0). Do not
allow any reference  to A1 as (1, 1) to creep into this code or it will ruin
everything. Referring to ('A', 1) is fine; there's a function for interpreting
cell references, but it uses the presence of a letter in the column as a cue that
the row should be counted from 1, so it can't handle counting columns from 1.
Some packages DO use numbers (from 1) instead of indices (from 0) internally. If
one of these packages is used, its returned values must be converted immediately.
YOU WON'T LIKE WHAT HAPPENS IF YOU FAIL TO MAINTAIN THIS DISCIPLINE.
"""

# OMFIT imports
try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

# Basic imports
import copy
import string
import numpy as np
import matplotlib
import re

# OMFIT class imports
from omfit_classes.omfit_mds import OMFITmdsValue
from omfit_classes.omfit_base import OMFITtmp, OMFITtree

# Import modules for interfacing with google sheets
try:
    # This is an optional dependency. If it isn't installed, we can't do much except
    # explain the problem.
    import pygsheets
except ImportError:
    pygsheets = None


setup_help = '''
How to prepare a Google Sheet for use with these tools:
-------------------------------------------------------
Most of these steps are described with screen shots in the pygsheets documentation
(use the service account version):
https://pygsheets.readthedocs.io/en/stable/authorization.html

1: Make a sheet in Google drive
2: Make a project at https://console.developers.google.com/
3: Add and enable both Google drive API and Google sheets API
4: Add credentials: create credentials for a service account. Give it the Project: editor role.
5. Add a JSON key to the service account, download it, and place it somewhere accessible to this script.
6: Note the service account's email address (the form is <SERVICE-NAME>@<PROJECT>.iam.gserviceaccount.com)
7: Share the sheet with the service account using the email address it gets (ending in iam.gserviceaccount.com)

Connection should then be possible if OMFITgoogleSheet or subclass is initialized
with the right keyfile and sheet_name.
'''

__all__ = ['OMFITgoogleSheet', 'OMFITexperimentTable']


# List of column letters (A, B, ... Z, AA, AB, ... AZ, BA, ... ZZ)
string_ascii_uppercase = list(map(chr, range(ord('A'), ord('Z') + 1)))
col_letters = [] + string_ascii_uppercase
for first_letter in string_ascii_uppercase:
    for second_letter in string_ascii_uppercase:
        col_letters.append(first_letter + second_letter)


def interpret_row_col(*args):
    """
    Interprets row number and column name or row index and column index and returns indices

    Either provide row number and column letter as in google sheet as in args = ('A', 1) or args = ('A1', )
    or provide the row and column indices starting from 0 as in args = (0, 0).

    That is, give the input the way google displays it or the way python tracks it internally. No mixing.

    Large column letters, like AA, AB, CX, etc. are supported up to ZZ.

    This function is useful because we are dealing with a mixture of methods that try to index arrays from 0 and
    methods that try to communicate with the spreadsheet and its cell addresses. It is very convenient to hand this
    thing a valid address from some source and let it figure out how to interpret it. Just make sure you DO NOT let
    cell references like (1, 1) for A1 into the code. Convert them on the way in into (0, 0) or ('A', 1).

    :param args: One or two input arguments.
        One argument should be an address like 'A1'
        Two arguments should be a column followed by a row.
            If the column is int, both it and the row as interpreted as indices starting from 0.
            If the column is in the list of recognized column letters (A, B, ... ZZ), the row is interpreted as a
                row number, starting from 1.
            So, args of (0, 0), ('A', 1) and 'A1' all mean the same thing.

    :return: (int, int, str)
        (column_index, row_index, address)
        The indices start from 0 and the address.
    """

    # Assign the input arguments to column, row, and address_in
    if len(args) == 1:
        address_in = args[0]
        column, row = re.findall(r'[A-Za-z]+|\d+', address_in)  # https://stackoverflow.com/a/28290501/6605826
        row = int(row)
    elif len(args) == 2:
        address_in = None
        column, row = args[0:2]
    else:
        raise ValueError('Too many arguments')

    # Decide what the indices for column and row are based on what's been given
    if isinstance(column, int):
        column_index = column
        row_index = int(row)
    elif column in col_letters:
        row_index = int(row) - 1
        column_index = int(np.where(col_letters == np.array(column))[0][0])
    else:
        raise ValueError(f'Could not understand column index {column}')

    # Check and then return results
    address = f'{col_letters[column_index]}{row_index+1}'
    if address_in is not None:
        assert address_in == address, 'Did not recover the right address'
    assert row_index >= 0, 'Row index must not be negative.'
    return column_index, row_index, address


def list_to_array_type(list_in, type_out=None, fill_value=np.NaN, int_truncation_allowed=False):
    """
    Forces a list to be an array with a specified type.

    Values that can't be forced to the type are replaced with a fill_value.
    Special: can be set to consider floats like 5.2 to be unrepresentable as ints instead of truncating them.

    :param list_in: list
        These would normally be the values read from a column or row.
        If formatted values are obtained, they will be strings, even if the strings just hold numbers.

    :param type_out: type
        This is just the type you would like each element to be in the output.

    :param fill_value: object matching type_out
        Elements that can't be forced to type_out will be replaced with this.

    :param int_truncation_allowed: bool
        Allow floats to be cast as int by truncation, so 26.945 becomes 26 instead of fill_value

    :return: array matching type_out
    """
    numeric_types = [int, float, complex]

    type_out = type_out or float
    try:
        type_out(fill_value)
    except Exception as excp:
        raise excp.__class__(f'fill_value {fill_value} is inconsistent with type_out {type_out}!')

    def try_type(x):
        """Tries to force x to be type and returns it; returns fill_value on failure"""
        # noinspection PyBroadException
        try:
            if type_out in numeric_types:
                the_thing = type_out(float(x))
            else:
                the_thing = type_out(x)
        except Exception as exc:
            printd(f'      failed to cast {type_out}({x}) because {exc}')
            return type_out(fill_value)
        else:
            if (not int_truncation_allowed) and (type_out == int) and (float(x) != the_thing):
                printd(f'    Casting to int would change {x} to {the_thing} and is not allowed. FAIL!')
                return type_out(fill_value)
            else:
                return the_thing

    # fromiter : https://stackoverflow.com/a/28526475/6605826
    return np.fromiter(map(try_type, list_in), type_out)


def operate_on_data(data, operation):
    """
    Handles mean/median/etc. operations for clients such as OMFITexperimentTable(...).write_mds_data_to_table()

    :param data: 1D float array
        Data to process

    :param operation: str
        Description of operation

    :return: float
        Result of operation on data
    """
    if operation == 'mean':
        value = np.nanmean(data)
    elif operation == 'median':
        value = np.nanmedian(data)
    elif operation == 'max':
        value = np.nanmax(data)
    elif operation == 'min':
        value = np.nanmin(data)
    else:
        raise ValueError(f'Unrecognized operation passed to operate_on_data(): {operation}')
    return value


def get_definition_from_docstring(theobject, return_kw=False):
    """
    Reads example code from the docstring of a class or function and uses it to define some variables

    :param theobject: class or function
        The docstring should contain statements like '>>> the_variable = the_value'.
        They can be multi-line.
        Each statement should be separated from others by lines not starting with >>> .

    :param return_kw: bool
        Instead of evaluating expressions as written (e.g. var = ClassName(**kw)),
        replace class or method name with dict and append _kw to var, to instead get
        var_kw = dict(**kw). This will fail if the statements in the docstring aren't
        don't follow the form a = b(keywords=values, keywords2=values2, ...).

    :return: dict
        Keys are the variables being defined (on the left side of the first =)
        Values are evaluated expressions taken from the right side of the first =.
    """
    fn_doc = theobject.__doc__.split('\n')
    is_example = np.array([doc_line.strip().startswith('>>> ') for doc_line in fn_doc])
    exmp_changes = np.append(0, np.diff(is_example.astype(int)))
    starts = np.where(exmp_changes == 1)[0]
    ends = np.where(exmp_changes == -1)[0]
    docstring_definitions = {}
    for start, end in zip(starts, ends):
        var_name = fn_doc[start].split('=')[0].strip().split('>>> ')[1]
        def_start = '='.join(fn_doc[start].split('=')[1:]).strip()
        def_rest = [d.split('>>> ')[1] for d in fn_doc[start + 1 : end]]
        definition = '\n'.join([def_start] + def_rest)
        if return_kw:
            # noinspection PyBroadException
            try:
                class_or_func_name = definition.split('(')[0]
                replaced_with_dict = 'dict' + definition[len(class_or_func_name) :]
                docstring_definitions[var_name + '_kw'] = eval(replaced_with_dict)
            except Exception:
                docstring_definitions[var_name + '_kw'] = None
        else:
            docstring_definitions[var_name] = eval(definition)
    return docstring_definitions


class OMFITgoogleSheetSetupError(ValueError):
    pass


class OMFITgoogleSheetBadKeyfileError(OMFITgoogleSheetSetupError):
    pass


class OMFITexperimentTableSetupError(OMFITgoogleSheetSetupError):
    pass


class OMFITgoogleSheet(OMFITtree):
    """
    Connects to Google sheets and provides some convenience features

    * Lookup columns by column header name instead of column number

    * Local caching to allow repeated read operations without too many connections to the remote sheet.

    * Throttle to avoid throwing an exception due to hitting Google's rate limiter,
      or pygsheet's heavy-handed rate limit obedience protocol.

    An example sheet that is compatible with the assumptions made by this class may be found here:
    https://docs.google.com/spreadsheets/d/1MJ8cFjFZ2pkt4OHWWIciWT3sM2hSADqzG78NDiqKgkU/edit?usp=sharing

    A sample call that should work to start up an OMFITgoogleSheet instance is:
    >>> gsheet = OMFITgoogleSheet(
    >>>     keyfile=os.sep.join([OMFITsrc, '..', 'samples', 'omfit-test-gsheet_key.json']),
    >>>     sheet_name='OmfitDataSheetTestsheet',
    >>>     subsheet_name='Sheet1',  # Default: lookup first subsheet
    >>>     column_header_row_idx=5,  # Default: try to guess
    >>>     units_row_idx=6,  # Default: None
    >>>     data_start_row_idx=7,  # Default: header row + 1 (no units row) or + 2 (if units row specified)
    >>> )
    This call should connect to the example sheet. This is more than an example; this is a functional call
    that is read out of the docstring by the regression test and testing will fail if it doesn't work properly.
    """

    def __init__(
        self,
        filename='',
        keyfile=None,
        sheet_name=None,
        subsheet_name=None,
        column_header_row_idx=None,
        column_header_row_number=None,
        units_row_idx=None,
        units_row_number=None,
        data_start_row_idx=None,
        data_start_row_number=None,
        **kw,
    ):
        """
        :param filename: str
            Not used, but apparently required when subclassing from OMFITtree.

        :param keyfile: str or dict-like
            Filename with path of the file with authentication information,
            or dict-like object with the parsed file contents (OMFITjson should work well).
            See setup_help for help setting this up.

        :param sheet_name: str
            Name of the Google sheets file/object/whatever to access

        :param subsheet_name: str
            Sheet within the sheet (the tabs at the bottom). Defaults to the first tab.

        :param column_header_row_idx: int
            Index (from 0) of the row with column headers. If not specified, we will try to guess for you.
            Indices like this are stored internally.

        :param column_header_row_number: int
            Number (from 1) of the row with column headers, as it appears in the sheet.
            Ignored if column_header_row_idx is specified. If neither is specified, we will try to guess for you.
            This will be converted into an index (from 0) for internal use.

        :param units_row_idx: int or None
            Index (from 0) of the row with the units, if there is one, or None if there isn't a row for units.

        :param units_row_number: int or None
            Number (from 1) of the units row. See description of column_header_row_number.

        :param data_start_row_idx: int
            Index (from 0) of the row where data start.
            Defaults to column_header_row + 1 if units_row is None or +2 if units_row is not None.

        :param data_start_row_idx: int
            Number (from 1) of the first row of data after the header.
            See description of column_header_row_number.

        :param kw: additional keywords passed to super class's __init__
        """
        self.OMFITproperties = {}
        super().__init__(filename=filename, **kw)

        # Make a scratch space for objects that don't need to be / can't be saved
        self['scratch'] = OMFITtmp()
        self['scratch']['client'] = None
        self['scratch']['sheet_connection'] = None

        # Cache setup
        self['__cache__'] = OMFITtmp()
        self['cache_expiration'] = 30  # seconds

        # Throttle setup
        self['previous_write_times'] = np.array([])
        self['throttle_interval_s'] = 100.0
        self['throttle_requests_in_interval'] = 100

        # Interpret input keywords
        self['keyfile'] = keyfile
        self['sheet_name'] = sheet_name
        self['subsheet_name'] = subsheet_name
        if column_header_row_idx is None and column_header_row_number is not None:
            column_header_row_idx = column_header_row_number - 1
        self['column_header_row'] = column_header_row_idx
        if units_row_idx is None and units_row_number is not None:
            units_row_idx = units_row_number - 1
        self['units_row'] = units_row_idx
        if data_start_row_idx is None and data_start_row_number is not None:
            data_start_row_idx = data_start_row_number - 1
        self['data_start_row'] = data_start_row_idx

    def __tree_repr__(self):
        if pygsheets is None:
            return 'Import failed: pygsheets. No connection possible.', []
        if isinstance(self['scratch'].get('sheet_connection', None), pygsheets.Worksheet):
            connection_status = 'connected'
        else:
            connection_status = '(NOT connected)'
        subsheet = self.get('subsheet_name', None)
        if subsheet is None:
            subsheet = 'TBD'
        description = (
            f"{self.get('sheet_name', '')}/{subsheet} {connection_status} "
            f"{self['column_header_row']}|{self['units_row']}|{self['data_start_row']}"
        )
        return description, []

    def __call__(self):
        """This seems like a useful thing to do"""
        self.connect()

    def record_request(self):
        """Logs a request to the Google API so the rate can be avoided by local throttling"""
        self['previous_write_times'] = np.append(self['previous_write_times'], time.time())

    def self_check(self, essential_only=False):
        """
        Makes sure setup is acceptable and raises AssertionError otherwise

        :param essential_only: bool
            Skip checks that aren't essential to initializing the class and its connection.
            This avoids spamming warnings about stuff that might get resolved later in setup.
        """
        problems = []
        issues = []

        # noinspection PyBroadException
        try:
            # Essential checks
            if self['keyfile'] is None:
                problems += ['keyfile missing: Need filename (with path) of the key for accessing Google sheets.']
            if self['sheet_name'] is None:
                problems += ['sheet_name missing: Please specify the name of the sheet to access.']

            # More checks
            if not essential_only:
                if self['column_header_row'] is None:
                    issues += ['column_header_row missing: required for most operations.']
        except Exception as excp:
            problems += [f'Encountered exception while trying to check stuff: {excp}']
        # Report problems
        if len(issues):
            fissues = '\n'.join(issues)
            print(f'Warning: there were {len(issues)} issues that may prevent full functionality:\n{fissues}')
        fproblems = '\n'.join(problems)
        if problems:
            self.help()
        assert len(problems) == 0, f'There were some setup problems:\n{fproblems}'

    def authorize_client(self):
        """
        Deal with keyfile being allowed to be either a string or a dict-like

        :return: pygsheets.client.Client instance
        """
        if is_string(self['keyfile']):
            return pygsheets.authorize(service_account_file=self['keyfile'])
        elif isinstance(self['keyfile'], dict):
            from google.oauth2 import service_account

            clean_keyfile = dict(self['keyfile'])  # Things get weird if keyfile is an OMFITjson; simplify to dict
            scopes = ('https://www.googleapis.com/auth/spreadsheets', 'https://www.googleapis.com/auth/drive')
            creds = service_account.Credentials.from_service_account_info(clean_keyfile, scopes=scopes)
            return pygsheets.authorize(custom_credentials=creds)
        else:
            raise OMFITgoogleSheetBadKeyfileError(
                f"Could not use keyfile of type {type(self['keyfile'])}; should be either a str or dict-like"
            )

    def connect(self, attempt=0):
        """Establishes connection to google sheet"""
        self.self_check(essential_only=True)

        if pygsheets is None:
            exc = AssertionError("Failed to import required packages: can't connect to google.")
            self['scratch']['sheet_connection'] = exc
            printe('Unable to connect because of missing packages')
            return

        print('Connecting to google sheets...')
        try:
            self['scratch']['client'] = self.authorize_client()
            # .open() here is not the same as file.open(). Do not put in `with`.
            opened_client = self['scratch']['client'].open(self['sheet_name'])
            self['subsheet_name'] = self['subsheet_name'] or opened_client.worksheets()[0].title
            self['scratch']['sheet_connection'] = (
                self['scratch']['client'].open(self['sheet_name']).worksheet_by_title(self['subsheet_name'])
            )
        except Exception as exc:
            self['scratch']['sheet_connection'] = exc
            if attempt < 1:
                wait_time = 2.0
                printw(f'First attempt to connect failed; waiting {wait_time} seconds and trying again. ({exc})')
                time.sleep(wait_time)
                self.connect(attempt=attempt + 1)
            else:
                printe(f'Failed to connect ({exc})')
                self['scratch']['sheet_connection'] = exc
        else:
            self.record_request()
            self.post_connect()

    def sheet_connection(self):
        """
        Attempt to connect on demand and return a Worksheet object

        :return: Worksheet instance
        """
        if self['scratch'].get('sheet_connection', None) is None:
            self.connect()
        return self['scratch'].get('sheet_connection', None)

    def post_connect(self):
        """Fill in some handy information; usually desired after connecting"""
        if self['column_header_row'] is None:
            self.guess_header_row()
        default_data_row = self['column_header_row'] + 1 + (self['units_row'] is not None)
        self['data_start_row'] = self['data_start_row'] or default_data_row
        self.get_columns()

    @staticmethod
    def help():
        """Prints help"""
        print(setup_help)

    def direct_write(self, value, *args):
        """
        Writes data to a single cell in the remote sheet

        :param value: numeric or string
            This will be written to the cell

        :param args: address information, such as (column, row) or address string.
            Examples: ('A', 1), (0, 0), 'A1'.
            See interpret_row_col() for more information about how to specify row and column.
        """
        from omas.omas_core import force_imas_type

        column_index, row_index, address = interpret_row_col(*args)
        self.sheet_connection().update_value(address, force_imas_type(value))
        self.record_request()
        self['__cache__']['changed'] = True

    def direct_read(self, *args, formatted=False):
        """
        Reads data from a single cell in the remote sheet

        :param args: passed to interpret_row_col
            Provide address information here, like ('A', 1), (0, 0), or 'A1'

        :param formatted: bool
            Get the formatted value (floats will be rounded according to sheet display settings)

        :return: str or number
            If formatted = True: str
                Might contain something that can be trivial cast as int or float, but it's always in a string
            If formatted = False: str or number
                With formatting off, numbers may be returned
        """
        column_index, row_index, address = interpret_row_col(*args)
        self.obey_rate_limit()
        cell = self.sheet_connection().cell(address)
        self.record_request()
        if formatted:
            return cell.value
        else:
            return cell.value_unformatted

    def time_until_throttle_okay(self):
        """Checks history of previous requests and decides how long we must wait before making a new request"""
        # pygsheets has a seconds_per_quota setting, but all it does is wait that long if it hits an exception.
        # It won't be smart and figure out that you only have to wait N seconds to be under the rate again.
        t0 = time.time()
        self['scratch']['t0'] = t0
        recent_requests = np.sum((t0 - self['previous_write_times']) < self['throttle_interval_s'])
        if recent_requests < self['throttle_requests_in_interval']:
            printd(f'Number of recent requests is only {recent_requests}, so there is no need to wait.')
            wait_time = 0
        else:
            while np.max((t0 - self['previous_write_times']) - self['throttle_interval_s']) > 0:
                # Remove old values that are past timeout
                self['previous_write_times'] = self['previous_write_times'][1:]
            # How long did that stuff take? I don't know, let's update t0 just in case we wasted time
            t0 = time.time()
            since = t0 - self['previous_write_times']
            since_interval = since - self['throttle_interval_s']
            possible_wait_times = np.arange(self['throttle_interval_s'] + 1)
            since_interval_after_wait = since_interval[:, np.newaxis] + possible_wait_times[np.newaxis, :]
            in_interval_after_wait = np.sum(since_interval_after_wait <= 0, axis=0)
            acceptable = in_interval_after_wait < self['throttle_requests_in_interval']
            wait_time = possible_wait_times[np.where(acceptable)[0][0]]
            self['scratch']['since'] = since_interval
            self['scratch']['since_interval'] = since_interval
            self['scratch']['since_interval_after_wait'] = since_interval_after_wait
            self['scratch']['in_interval_after_wait'] = in_interval_after_wait
            self['scratch']['acceptable'] = acceptable

        return wait_time

    def obey_rate_limit(self):
        """Waits as needed to avoid rate limit"""
        wait_time = self.time_until_throttle_okay()
        if wait_time > 0:
            printw('Connections are being throttled to avoid an error due to too many requests to Google')
            print(f'Waiting {wait_time} seconds before making another request...')
            time.sleep(wait_time)

    def throttled_write(self, value, *args):
        """
        Checks whether it's okay to make a request and waits if necessary to avoid an error

        :param value: numeric or string

        :param args: arguments to pass to direct_read()
        """
        self.obey_rate_limit()
        self.direct_write(value, *args)

    def throttled_read(self, *args, **kw):
        """
        Checks whether it's okay to make a request and waits if necessary to avoid an error

        :param args: arguments to pass to direct_read()

        :param kw: keywords to pass to direct_read()

        :return: value returned from direct_read()
        """
        self.obey_rate_limit()
        self.direct_read(*args, **kw)

    def guess_header_row(self):
        """
        Tries to determine which row holds column headers based on the presence of likely column names
        """
        from collections import OrderedDict

        print(
            'Attempting to guess which row has the column headers. '
            'This process is slow and it is better to supply this information at initialization.'
        )
        candidates = dict()
        occurrences = 0

        keywords = OrderedDict(shot=10, device=0.1, id=1, comments=0.01)

        for keyword, value in keywords.items():
            print(f'Checking for {keyword}')
            self.obey_rate_limit()
            cells = self.sheet_connection().find(keyword, matchEntireCell=True, matchCase=False)
            self.record_request()
            if len(cells) == 1:
                row = cells[0].row
                candidates[row] = candidates.get(row, 0) + value
            if len(candidates) > 0:
                best_row = max(candidates, key=candidates.get)
                high_score = candidates[best_row]
                occurrences = sum([v == high_score for v in candidates.values()])
                if occurrences == 1:
                    self['column_header_row'] = best_row - 1
                    print(f'Assigned column_header_row = {best_row - 1}')
                    return

        if len(candidates) == 0:
            print('Could not find any of the keywords that would help guess which row had the headers.')
            print('Failed to guess header row.')
            return

        if occurrences > 1:
            print('More than one row seemed like it might have column headers. No rule for dealing with this.')
            print('Failed to guess header row.')
            return

    def get_columns(self):
        """Gets a list of column names from the header row"""
        if self['column_header_row'] is not None:
            self.obey_rate_limit()
            self['columns'] = self.sheet_connection().get_row(self['column_header_row'] + 1)
            self.record_request()
        else:
            print('Cannot get columns until column_header_row is specified')

    def find_column(self, column_name, quiet=False):
        """
        Finds the column index corresponding to a column name

        :param column_name: str
            Column header name (should be in row given by column_header_row)

        :param quiet: bool
            No printing

        :return: int (success) or None (failed to find column)
            Column index (from 0; not number from 1)
        """
        indices = np.where(np.array(self['columns']) == column_name)[0]
        if len(indices) > 1:
            print(f'WARNING! More than one column matches {column_name}! They are: {indices}')
            return [idx for idx in indices]
        elif len(indices) == 0:
            if not quiet:
                print(f"Error! Column {repr(column_name)} not found in {self['columns']}")
            return None
        else:
            return int(indices[0])

    def update_cache(self, force=True):
        """
        Updates the cached representation if forced or if it seems like it's a good idea

        If it's been a while since the cache was updated, do another update.
        If we know there was a write operation since the last update, do another update.
        If the cache is missing, grab the data and make the cache.
        If force=True, do another update.

        :param force: bool
            Force an update, even if other indicators would lead to skipping the update
        """
        dt = time.time() - self['__cache__'].get('time', 0)
        update_needed = (
            force
            or self['__cache__'].get('changed', True)
            or (dt > self['cache_expiration'])
            or (self['__cache__'].get('values', None) is None)
        )
        if not update_needed:
            return
        self.obey_rate_limit()
        self['__cache__']['values'] = self.sheet_connection().get_all_values(
            returnas='matrix',
            majdim='COLS',
            include_tailing_empty=True,
            include_tailing_empty_rows=True,
            value_render='UNFORMATTED_VALUE',
            date_time_render_option='SERIAL_NUMBER',
        )
        self.record_request()
        self['__cache__']['time'] = time.time()
        self['__cache__']['changed'] = False

    def get_column(self, column, force_update_cache=False, disable_cache=False):
        """
        Gets data from a column

        By default, the local cache will be checked & updated if needed and then data will be read from the cache.
        Caching can be disabled, which will result in more connections to the remote sheet. Some other methods have
        been programmed assuming that local caching done here will save them from otherwise inefficient layout of
        calls to this function.

        :param column: int or str
            Column index (from 0) or column letter (like 'A', 'B', ... 'ZZ')

        :param force_update_cache: bool
            Force the cache to update before reading

        :param disable_cache: bool
            Don't go through the local cache; read the remote column directly

        :return: list
            Data from the column
        """
        col_idx = interpret_row_col(column, 1)[0]
        if disable_cache:
            self.obey_rate_limit()
            tmp = self.sheet_connection().get_col(col_idx + 1)
            self.record_request()
            return tmp
        self.update_cache(force=force_update_cache)
        return self['__cache__']['values'][col_idx]

    def get_column_by_name(self, column_name, **kw):
        """
        Gets data from a column by name

        :param column_name: str
            Name of the column in the header row

        :param kw: additional parameters passed to get_column

        :return: list
            Data from the column
        """
        return self.get_column(self.find_column(column_name), **kw)

    def get_row(self, row_idx, force_update_cache=False, disable_cache=False):
        """
        Gets data from a row

        By default, cached results are used after updating the local cache as needed (see get_column for details).

        :param row_idx: int
            Row index (from 0)

        :param force_update_cache: bool
            Force the cache to update before reading

        :param disable_cache: bool
            Don't go through the local cache; read the remote row directly

        :return: array
            Data from the row
        """
        if disable_cache:
            self.obey_rate_limit()
            tmp = self.sheet_connection().get_row(row_idx + 1)
            self.record_request()
            return np.array(tmp)
        self.update_cache(force=force_update_cache)
        return np.array(self['__cache__']['values']).T[row_idx]

    # End of class OMFITgoogleSheet


class OMFITexperimentTable(OMFITgoogleSheet):
    """
    Extends OMFITgoogleSheet by assuming each row corresponds to a shot, allowing more features to be provided.

    This is less general (now Shot must exist), but it may be more convenient
    to look up data by shot this way. With more assumptions about the structure
    of the sheet, more methods can be implemented.

    Many methods go further to assume that there is a column named 'Device' and
    that there are two columns that can be used to determine a time range. Parts
    of the class should work on sheets without this structure, but others will
    fail.

    An example sheet that is compatible with the assumptions made by this class may be found here:
    https://docs.google.com/spreadsheets/d/1MJ8cFjFZ2pkt4OHWWIciWT3sM2hSADqzG78NDiqKgkU/edit?usp=sharing

    A sample call that should work to start up an OMFITgoogleSheet instance is:
    >>> xtable = OMFITexperimentTable(
    >>>     keyfile=os.sep.join([OMFITsrc, '..', 'samples', 'omfit-test-gsheet_key.json']),
    >>>     sheet_name='OmfitDataSheetTestsheet',
    >>>     subsheet_name='Sheet1',  # Default: lookup first subsheet
    >>>     column_header_row_idx=5,  # Default: try to guess
    >>>     units_row_idx=6,  # Default: search for units row
    >>>     data_start_row_idx=7,  # Default: find first number after the header row in the shot column
    >>> )
    This call should connect to the example sheet. These data are used in regression tests and should be updated
    If the test sheet is changed.
    """

    def __init__(self, *args, **kw):
        """
        :param kw: passed to parent class: OMFITgoogleSheet
        """
        self.OMFITproperites = {}
        data_start_row_idx = kw.get('data_start_row_idx', None)
        if data_start_row_idx is None and kw.get('data_start_row_number', None) is not None:
            data_start_row_idx = kw['data_start_row_number'] - 1
        self.input_data_start_row_idx = data_start_row_idx
        super().__init__(*args, **kw)
        self.shots_raw = None

    def post_connect(self):
        """
        Do some inspection on the sheet after connecting to it
        """
        super().post_connect()
        shot_col_idx = self.find_column('Shot')
        if shot_col_idx is None:
            raise OMFITexperimentTableSetupError(
                f"The sheet must contain a \"Shot\" column:\n"
                f"one cell in the header row must contain the exact text \"Shot\".\n"
                f"Was the header row specified correctly?\n"
                f"column_header_row_idx (currently {self['column_header_row']}) counts from 0,\n"
                f"while column_header_row_number (currently {self['column_header_row'] + 1}) counts from 1. "
                f"({self['column_header_row'] + 1} is what appears in Google's interface)"
            )
        self.shots_raw = self.get_column(shot_col_idx)
        self['units_row'] = self['units_row'] or self.find_units_row()
        self['data_start_row'] = self.input_data_start_row_idx or self.find_data_start_row()
        self['shots'] = self.shots_raw[self['data_start_row'] :]
        # Ensure there are no trailing blank entries in the list of shots
        notblank = np.where([len(str(shot)) > 0 for shot in self['shots']])[0]
        if len(notblank):
            self['shots'] = self['shots'][: notblank[-1] + 1]

    def self_check(self, essential_only=False):
        # Add additional checks here
        super().self_check(essential_only=essential_only)

    def guess_header_row(self):
        """
        Tries to determine which row holds column headers

        This is a simplification of the parent class's methods, because here we explicitly require a 'Shot' column.
        It should be faster and more reliable this way.
        """
        print('Attempting to guess which row has the column headers.')
        key_value = 'Shot'
        self.obey_rate_limit()
        cells = self.sheet_connection().find(key_value, matchEntireCell=True)
        self.record_request()

        assert len(cells) == 1, 'There should only be one cell that contains exactly "Shot"'
        self['column_header_row'] = cells[0].row - 1
        print(f"Assigned column_header_row = {self['column_header_row']}")

    def find_units_row(self):
        """
        Recommended: put a row of units under the column headers (data won't start on the next row after headers)

        Were recommendations followed? See if 1st row under headers has "units" under a
        column named Shot, id, or ID. In this case, data start on the row after units.

        This function is easier to implement under OMFITexperimentTable instead of
        OMFITgoogleSheet because we are allowed to make additional assumptions about the
        content of the table.

        If there is no cell containing just 'units', 'Units', or 'UNITS', but you do
        have a row for units, you should specify units_row at init.

        :return: int
            Index (from 0) of the row with units
        """
        if self.shots_raw is None:
            self.connect()

        # If one of the entries under "Shot" is just 'units', this is easy. Just find it.
        units_names = ['units', 'Units', 'UNITS']
        for uu in units_names:
            if uu in self.shots_raw:
                return np.where(np.array(self.shots_raw) == uu)[0][0]

        # So 'units' wasn't written under 'Shot'. Maybe under 'id' or something?
        col_names = ['id', 'ID', 'Device']
        for col_name in col_names:
            col_idx = self.find_column(col_name, quiet=True)
            if col_idx is None:
                continue
            self.obey_rate_limit()
            raw_col_values = self.sheet_connection().get_col(col_idx + 1)
            self.record_request()
            for uu in units_names:
                if uu in raw_col_values:
                    return np.where(np.array(raw_col_values) == uu)[0][0]

        # Still didn't find it? Maybe a "units" row wasn't included.
        return None

    def find_data_start_row(self):
        """
        Determine the first row of data.

        Allowing this to be different from header + 1 means meta data can be inserted under the column headers.

        Since this class imposes the requirement that the "Shot" column exists & has shot numbers, we can find the
        first valid number in that column, after the header row, and be safe in assuming that the data start there.

        To allow for nonstandard configurations of blank cells, please specify data_start_row manually in init.

        :return: int
            Index (from 0) of the row where data start.
        """
        if self.shots_raw is None:
            self.connect()
        for i, val in enumerate(self.shots_raw[self['column_header_row'] :]):
            if i == 0:
                continue  # Just skip the header row
            try:
                int(val)
            except ValueError:
                pass
            else:
                return self['column_header_row'] + i
        # Didn't find a value that could be cast as int after the header
        return self['column_header_row'] + (self['units_row'] is not None) + 1

    def extract_column(self, column, pad_with='', force_type=None, fill_value=np.NaN, fill_if_not_found='not found'):
        """
        Downloads data for a given column and returns them with units

        :param column: str or int
            str: Name of the column; index will be looked up
            int: Index of the column, from 0. Name of the column can be looked up easily.

        :param pad_with: str
            Value to fill in to extend the column if it is too short.
            Truncation of results to the last non-empty cell is possible, meaning columns
            can have inconsistent lengths if some have more empty cells than others.

        :param force_type: type [optional]
            If provided, pass col_values through list_to_array_type() to force them to match force_type.

        :param fill_value: object
            Passed to list_to_array_type(), if used; used to fill in cells that can't be forced to force_type.
            It probably would make sense if this were the same as pad_with, but it doesn't have to be.
            It does have to be of type force_type, though.

        :param fill_if_not_found: str
            Value to fill in if data are not found, such as if column_name is None.

        :return: (array, str or None)
            values in the column, padded or cut to match the number of shots
                replaced by `fill_if_not_found` in case of bad column specifications
                replaced by `fill_value` for individual cells that can't be forced to type `force_type`
                extended by `pad_with` as needed
            units, if available or None otherwise
        """
        if 'shots' not in self:
            self.connect()
        ns = len(self['shots'])

        if is_string(column):
            column_name = column
            column_idx = self.find_column(column_name)
        elif isinstance(column, (int, np.integer)):
            column_idx = column
            column_name = self['columns'][column_idx]
        else:
            raise ValueError('Invalid column specification; should be string (column_name) or int (column_index)')

        if column_name is None:
            return np.array([fill_if_not_found] * ns), None

        col_values0 = self.get_column(column_idx)
        deficit = (ns + self['data_start_row']) - len(col_values0)
        if deficit > 0:
            col_values0 += [pad_with] * deficit
        elif deficit < 0:
            col_values0 = col_values0[:deficit]
        if self.get('units_row', None) is not None:
            units = col_values0[self['units_row']]
        else:
            units = None
        col_values = col_values0[self['data_start_row'] :]
        # print(f'units of {column_name} are listed as {repr(units)}')
        if force_type is not None:
            col_values = list_to_array_type(col_values, type_out=force_type, fill_value=fill_value)
        else:
            col_values = np.array(col_values)
        return col_values, units

    def interpret_timing(self, **signal):
        """
        Interprets timing information in the signal dictionary for write_mds_data_to_table().

        Either tmin and tmax or time and dt are required to be in the dictionary as keys.
        See write_mds_data_to_table() documentation for more information.

        :param signal: Just pass in one of the dictionaries in the list of signals.

        :return: (float array, float array)
            Array of start times and array of end times in ms. One entry per row.
        """
        tmin = signal.get('tmin', None)
        tmax = signal.get('tmax', None)
        t = signal.get('time', None)
        dt = signal.get('dt', None)
        if tmin is not None and tmax is not None:
            if is_string(tmin):
                tmin_name = tmin
                tmin, _ = self.extract_column(tmin, force_type=float)
            else:
                tmin_name = None
                tmin = np.zeros(len(self['shots'])) + float(tmin)
            if is_string(tmax):
                tmax_name = tmax
                tmax, _ = self.extract_column(tmax, force_type=float)
            else:
                tmax_name = None
                tmax = np.zeros(len(self['shots'])) + float(tmax)
            print(f'Got timing from tmin (column {tmin_name}) and tmax (column {tmax_name})')
        elif t is not None and dt is not None:
            if is_string(t):
                time_name = t
                t, _ = self.extract_column(t, force_type=float)
            else:
                time_name = None
                t = np.zeros(len(self['shots'])) + float(t)
            if is_string(dt):
                dt_name = dt
                dt, _ = self.extract_column(dt, force_type=float)
            else:
                dt_name = None
                dt = np.zeros(len(self['shots'])) + float(dt)
            tmin = t - dt
            tmax = t + dt
            print(f'Got timing from time (column {time_name}) and dt (column {dt_name})')
        else:
            raise ValueError('Need a valid pair of tmin & tmax or time & dt')
        return tmin, tmax

    def interpret_signal(self, i, device_servers, **signal):
        """
        Interprets signal specification data for write_mds_data_to_table()

        :param i: int
            Row / shot index

        :param device_servers: dict-like
            Dictionary listing alternative servers for different devices, like {'EAST': 'EAST_US'}

        :param signal: Just pass in one of the dictionaries in the list of signals.

        :return: tuple containing:
            server: str
            tree: str or None
            tdi: str
            factor: float
            tfactor: float
            device: str
        """
        devices = self.extract_column('Device')[0]
        column_name = signal['column_name']  # str, required
        server = device_servers.get(devices[i], devices[i])
        # Device-specific TDIs and treenames are possible if these are set as dictionaries.
        # Universal TDIs and treenames are easy; just don't put in a dictionary.
        # treename of None means PTDATA (d3d only).
        # TDI of None means re-use column_name (so you can put pointnames in as column_names and type this in once)

        def get_the_thing(name, default):
            info = signal.get(name, default)
            if isinstance(info, dict):
                thingy = info.get(devices[i], default)
            else:
                thingy = info
            return thingy

        tree = get_the_thing('treename', None)
        tdi = get_the_thing('TDI', column_name)
        tdi = tdi or column_name
        factor = get_the_thing('factor', 1.0)
        tfactor = get_the_thing('tfactor', 1.0)

        return server, tree, tdi, factor, tfactor, devices[i]

    def write_mds_data_to_table(self, signals, overwrite=False, device_servers=None):
        """
        Gets MDSplus data for a list of signals, performs operations on specified time ranges, & saves results.

        This sample signal can be used as an instructional example or as an input for testing
        >>> sample_signal_request = dict(
        >>>     column_name='density',  # The sheet must have this column (happens to be a valid d3d ptname)
        >>>     TDI=dict(EAST=r'\dfsdev'),  # d3d's TDI isn't listed b/c TDI defaults to column_name
        >>>     treename=dict(EAST='PCS_EAST'),  # DIII-D isn't listed b/c default is None, which goes to PTDATA
        >>>     factor={'DIII-D': 1e-13},  # This particular EAST pointname doesn't require a factor; defaults to 1
        >>>     tfactor=dict(EAST=1e3),  # d3d times are already in ms
        >>>     tmin='t min',  # str means read timing values from column w header exactly matching this string
        >>>     tmax='t max',
        >>> )
        This sample should work with the example/test sheet.

        If the sheet in question had ONLY DIII-D data, the same signal request could be accomplished via:
        >>> simple_sample_signal_request = dict(column_name='density', factor=1e-13, tmin='t min', tmax='t max')
        This sample won't work with the test sheet; it will fail on the EAST shots.
        We are exploiting the shortcut that we've used a valid pointname (valid for d3d at least)
        as the column header. If you want fancy column names, you have to specify TDIs.
        We are also relying on the defaults working for this case.

        The sample code in this docstring is interpreted and used by the regression test, so don't break it.
        Separate different samples with non-example lines (that don't start with >>>)

        :param signals: list of dict-likes
            Each item in this list should be a dict-like that contains information needed to fetch & process data.
            SIGNAL SPECIFICATION
            - column_name: str (hard requirement)
                Name of the column within the sheet
            - TDI: None, str, or dict-like (optional if default behavior (reuse column_name) is okay)
                None: reuse column_name as the TDI for every row
                str: use this str as the pointname/TDI for every row
                dict-like: keys should be devices. Each device can have its own pointname/TDI.
                The sheet must have a Device row. If a device is missing, it will inherit the column_name.
            - treename: None, str, or dict-like (optional if default behavior (d3d ptdata) is okay)
                None: PTDATA (DIII-D only)
                str: use this string as the treename for every row
                dict-like: device specific treenames. The sheet must have a Device row.
            - factor: float or dict-like (defaults to 1.0)
                float: multiply results by this number before writing
                dict-like: multiply results by a device specific number before writing (unspecified devices get 1)
            - tfactor: float or dict-like (defaults to 1.0)
                float: multiply times by this number to get ms
                dict-like: each device gets a different factor used to put times in ms (unspecified devices get 1)

            PROCESSING
            - operation: str (defaults to 'mean')
                Operation to perform on the gathered data. Options are:
                - 'mean'
                - 'median'
                - 'max'
                - 'min'
            - tmin: float or str
                Start of the time range in ms; used for processing operations like average.
                Must be paired with tmax.
                A usable tmin/tmax pair takes precedence over time+dt.
                A float is used directly. A string triggers lookup of a column in the sheet; then every row gets
                its own number determined by its entry in the specified column.
            - tmax: float or str
                End of the time range in ms. Must be paired with tmin.
            - time: float or str
                Center of a time range in ms. Ignored if tmin and tmax are supplied. Must be paired with dt.
            - dt: float or str
                Half-width of time window in ms. Must be paired with time.

        :param overwrite: bool
            Update the target cell even if it's not empty?

        :param device_servers: dict-like [optional]
            Provide alternative MDS servers for some devices. A common entry might be {'EAST': 'EAST_US'}
            to use the 'EAST_US' (eastdata.gat.com) server with the 'EAST' device.
        """

        # I'm assuming that caching in get_column() will save me from inefficient organization of read operations
        # In order to be good without caching, this function should figure out all the data it needs and read them
        # all up-front. That is, it should build its own temporary cache instead of relying on external caching.
        # But doing that adds a lot of development work and increases the maintenance burden, so I'm going to rely
        # on external caching instead.

        dserv = device_servers or {}

        if self.get('shots', None) is None:
            self.connect()
        shots = self['shots']

        for signal in signals:
            column_name = signal['column_name']
            column_idx = self.find_column(column_name)
            column_values, column_units = self.extract_column(column_name)
            tmin, tmax = self.interpret_timing(**signal)
            for i, shot in enumerate(shots):
                if (not overwrite) and len(str(column_values[i])) > 0:
                    print(f'Cell {column_name} {shot} already has a value and overwrite is off; skipping')
                    continue
                server, tree, tdi, factor, tf, _ = self.interpret_signal(i, device_servers=dserv, **signal)
                mds = OMFITmdsValue(server, shot=shot, treename=tree, TDI=tdi)
                if not mds.check():
                    print(f'Bad MDS connection for {server}#{shot} {tree} / {tdi}; skipping')
                    continue
                sel = ((mds.dim_of(0) * tf) >= tmin[i]) & ((mds.dim_of(0) * tf) <= tmax[i])
                data = mds.data()[sel] * factor
                value = operate_on_data(data, signal.get('operation', 'mean'))
                self.throttled_write(value, column_idx, self['data_start_row'] + i)

    def write_efit_result(self, pointnames=None, **kw):
        r"""
        Wrapper for write_mds_data_to_table for quickly setting up EFIT signals.

        Assumes that device-specific variations on EFIT pointnames and  primary trees
        are easy to guess, and that you have named your columns to exactly  match EFIT
        pointnames in their "pure" form (no leading \).

        Basically, it can build the signal dictionaries for you given a list of
        pointnames and some timing instructions.

        Here is a set of sample keywords that could be passed to this function:
        >>> sample_kw = dict(
        >>>     pointnames=['betan'],
        >>>     tmin='t min',  # Can supply a constant float instead of a string
        >>>     tmax='t max',
        >>>     overwrite=True,
        >>>     device_servers=dict(EAST='EAST_US'),  # Probably only needed for EAST
        >>> )
        These should work in xtable.write_efit_result where xtable is an
        OMFITexperimentTable instance connected to a google sheet that contains columns
        with headers 'betan', 't min', and 't max'

        :param pointnames; list of strings
            Use names like betan, etc. This function will figure out whether you really need \betan instead.

        :param kw: more settings, including signal setup customizations.
            * MUST INCLUDE tmin & tmax OR time & dt!!!!!!!!!              <---- don't forget to include timing data
            * Optional signal customization: operation, factor
            * Remaining keywords (other than those listed so far) will be passed to write_mds_data_to_table
            * Do not include column_name, TDI, treename, or tfactor, as these are determined by this function.
                If you need this level of customization, just use write_mds_data_to_table() directly.
        """

        signal_details = {tmp: kw.pop(tmp, None) for tmp in ['tmin', 'tmax', 'time', 'dt', 'operation', 'factor']}
        for sd in list(signal_details):
            if signal_details[sd] is None:
                signal_details.pop(sd)
        signals = [
            dict(
                column_name=pointname,
                TDI={'DIII-D': pointname, 'EAST': rf'\{pointname}'},
                treename=dict(EAST='EFIT_EAST'),
                tfactor=dict(EAST=1e3),
                **signal_details,
            )
            for pointname in pointnames
        ]
        self.write_mds_data_to_table(signals, **kw)

    # End of class OMFITexperimentTable


############################################
if __name__ == '__main__':
    # Basic class import test
    if pygsheets is not None:
        test_classes_main_header()
        test_sheet_data = get_definition_from_docstring(OMFITexperimentTable, return_kw=True)['xtable_kw']
        et = OMFITexperimentTable(
            keyfile=os.sep.join([OMFITsrc, '..', 'samples', 'omfit-test-gsheet_key.json']),
            sheet_name=test_sheet_data['sheet_name'],
            column_header_row_idx=test_sheet_data['column_header_row_idx'],
            units_row_idx=test_sheet_data['units_row_idx'],
            data_start_row_idx=test_sheet_data['data_start_row_idx'],
        )



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii

__all__ = ['OMFITreviewplus']


class OMFITreviewplus(SortedDict, OMFITascii):
    def __init__(self, filename='', **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):

        with open(self.filename, 'r') as f:
            lines = f.readlines()
        self['data'] = data = {}
        self['style'] = style = {}
        k = -1
        kst = -1
        mode = 'data'
        for line in lines[4:]:
            # ================
            if mode == 'data' and line.startswith('----'):
                k = -1
                mode = 'style'
                continue
            # ----------------
            if line.strip() == '53':
                mode = 'data'
                k += 1
                continue
            # ----------------
            if line.strip() == '66':
                kst += 1
                mode = 'style'
                continue
            # ================
            if mode == 'style':
                if kst not in style:
                    style[kst] = {}
                d = line.split('=')
                key = d[0]
                value = '='.join(d[1:]).strip()
                try:
                    style[kst][key] = eval(value)
                except Exception:
                    pass
                    # print key,value
            # ----------------
            if mode == 'data':
                if line.startswith('YEXP'):
                    k += 1
                if k not in data:
                    data[k] = {}
                data[k]['__style__'] = kst
                d = line.split('=')
                key = d[0]
                value = '='.join(d[1:]).strip()
                try:
                    data[k][key] = eval(value)
                except Exception:
                    pass
                    # print key,value
